#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Contains calibration related functions """

__all__ = ["create_calibration", "study_as_df", "ASCCalibrator", "ASCGroupCalibrator", "utils"]

import glob
import os
import shutil
import subprocess
import sys
from os import path, makedirs
from time import sleep
from typing import Union, Sequence, Callable, Tuple

import optuna
import yaml

from . import utils
from .base import CalibratorBase, to_float
from .calib_asc import ASCCalibrator
from .calib_group_asc import ASCGroupCalibrator
from .utils import study_as_df
from .analysis import calc_mode_stats


class CalibrationSampler(optuna.samplers.BaseSampler):
    """ Run calibrators on obtained mode shares """

    def __init__(self, calibrators: Sequence[CalibratorBase]):
        self.calibrators = {}
        for c in calibrators:
            if c.name not in self.calibrators:
                self.calibrators[c.name] = c
            else:
                raise ValueError("Incompatible calibrators with same name occurred: %s (%s)" % (c.name, c))

    def infer_relative_search_space(self, study, trial):
        return {}

    def sample_relative(self, study, trial, search_space):
        # study._storage.set_trial_system_attr(trial._trial_id, "search_space", self._search_space)
        return {}

    def sample_independent(self, study, trial, param_name, param_distribution):

        prefix, _, param = param_name.partition("-")
        c: CalibratorBase = self.calibrators[prefix]

        mode = param.rpartition("-")[2] if "-" in param else param

        completed = utils.completed_trials(study)
        if len(completed) == 0:
            initial = to_float(c.sample_initial(param))

            if c.constraints is not None and mode in c.constraints:
                initial = c.constraints[mode](mode, initial)

            return initial

        last = completed[-1]
        last_param = last.params[param_name]

        step = to_float(c.update_step(param, last, completed))

        rate = 1.0
        if c.lr is not None:

            rate = c.lr(float(len(completed) + 1), param_name, step, trial, study)

            # rate of None or 0 would be invalid
            if not rate:
                rate = 1

            # numpy types need casting
            rate = to_float(rate)

        # Need to use storage directly, frozen trial does not save attributes
        study._storage.set_trial_user_attr(trial._trial_id, "%s_rate" % param_name, rate)
        study._storage.set_trial_user_attr(trial._trial_id, "%s_step" % param_name, step)

        last_param += rate * step

        # Call constraint if present
        if c.constraints is not None and mode in c.constraints:
            last_param = c.constraints[mode](mode, last_param)

        return last_param


def create_calibration(name: str, calibrate: Union[CalibratorBase, Sequence[CalibratorBase]],
                       jar: Union[str, os.PathLike], config: Union[str, os.PathLike],
                       args: Union[str, Callable] = "", jvm_args="",
                       transform_persons: Callable = None,
                       transform_trips: Callable = None,
                       chain_runs: Union[bool, int, Callable] = False,
                       storage: optuna.storages.RDBStorage = None,
                       custom_cli: Callable = None,
                       debug: bool = False
                       ) -> Tuple[optuna.Study, Callable]:
    """ Create or load an existing study for mode share calibration using asc values.

    This function returns the study and optimization objective as tuple. Which can be used like this:
        study.optimize(obj, 10)

    :param name: name of the study
    :param calibrate: (list) of calibrator instances
    :param jar: path to executable jar file of the scenario
    :param config: path to config file to run
    :param args: additional arguments to the executable jar, can also be a callback function
    :param jvm_args: additional jvm args
    :param transform_persons: callable to filter persons included in mode share
    :param transform_trips: callable to modify trips included in mode share
    :param chain_runs: automatically use the output plans of each run as input for the next, either True or number of iterations or callable
    :param storage: custom storage object to overwrite default sqlite backend
    :param custom_cli: the scenario is not a matsim application and needs a different command line syntax
    :param debug: enable debug output
    :return: tuple of study and optimization objective.
    """

    # Init with 0
    if isinstance(calibrate, CalibratorBase):
        calibrate = [calibrate]

    # Set some custom arguments to prevent known errors on NFS
    if storage is None:
        storage = optuna.storages.RDBStorage(url="sqlite:///%s.db" % name,
                                             engine_kwargs={"connect_args": {"timeout": 100},
                                                            "isolation_level": "AUTOCOMMIT"})

    if not os.access(jar, os.R_OK):
        raise ValueError("Can not access JAR File: %s" % jar)

    if not os.access(config, os.R_OK):
        raise ValueError("Can not access config File: %s" % config)

    study = optuna.create_study(
        study_name=name,
        storage=storage,
        load_if_exists=True,
        directions=["minimize"] * sum(c.num_targets for c in calibrate),
        sampler=CalibrationSampler(calibrate)
    )

    if not path.exists("params"):
        makedirs("params")

    if not path.exists("runs"):
        makedirs("runs")

    study.set_user_attr("calib", [c.name for c in calibrate])
    for c in calibrate:
        c.init_study(study)

    def f(trial):

        params_path = path.join("params", "run%d.yaml" % trial.number)
        params = {}

        for c in calibrate:
            prefix = c.name + "-"
            c.update_config(study, trial, prefix, params)

        with open(params_path, "w") as f:
            yaml.dump(params, f, sort_keys=False)

        run_dir = "runs/%03d" % trial.number

        if os.path.exists(run_dir):
            shutil.rmtree(run_dir)

        completed = utils.completed_trials(study)

        # Call args if necessary
        run_args = args(completed) if callable(args) else args

        if custom_cli:
            cmd = custom_cli(jvm_args, jar, config, params_path, run_dir, trial.number, run_args)
        else:
            cmd = "java %s -jar %s run --config %s --yaml %s --output %s --runId %03d %s" \
                  % (jvm_args, jar, config, params_path, run_dir, trial.number, run_args)

        # Max fix extra whitespaces in args
        cmd = cmd.strip()

        if chain_runs:

            out = None
            for t in reversed(completed):
                out = glob.glob("runs/%03d/*.output_plans.xml.gz" % t.number)
                if out:
                    out = path.abspath(out[0])
                    break

            if out:
                last_chained = study.user_attrs.get("last_chained_run", None)
                if (chain_runs is True or
                        (callable(chain_runs) and chain_runs(completed)) or
                        (type(chain_runs) == int and len(completed) % chain_runs == 0)):
                    cmd += " --config:plans.inputPlansFile=" + out
                    study.set_user_attr("last_chained_run", out)

                elif last_chained:
                    cmd += " --config:plans.inputPlansFile=" + last_chained

            else:
                print("No output plans for chaining runs found.")

        # Extra whitespaces will break argument parsing
        cmd = cmd.strip()

        print("Running cmd %s" % cmd)

        if os.name != 'nt':
            cmd = cmd.split(" ")

        p = subprocess.Popen(cmd,
                             stdout=sys.stdout if debug else subprocess.DEVNULL,
                             stderr=sys.stderr if debug else subprocess.DEVNULL)
        try:
            while p.poll() is None:
                sleep(1)

            if p.returncode != 0:
                print("The scenario could not be run properly and returned with an error code.", file=sys.stderr)
                if not debug:
                    print("Set debug=True and check the output for any errors.", file=sys.stderr)
                    print("Alternatively run the cmd from the log above manually and check for errors.", file=sys.stderr)

                raise Exception("Process returned with error code: %s." % p.returncode)
        finally:
            p.terminate()

        mode_stats = calc_mode_stats(run_dir,
                                     transform_persons=transform_persons,
                                     transform_trips=transform_trips)

        trial.set_user_attr("mode_stats", mode_stats.to_dict(orient="tight"))

        err = []
        for c in calibrate:
            e = c.calc_stats(trial, run_dir, transform_persons, transform_trips)
            err.extend(to_float(f) for f in e)

        return err

    return study, f


# -*- coding: utf-8 -*-

import os

import numpy as np
import pandas as pd

from .. import *

# Has households, persons and trips
INPUT_FILES = 3


def is_format(f: os.DirEntry):
    fp = f.name
    if not f.path.endswith(".csv"):
        return False
    if "MiD" in f.name:
        return False

    return "_HH" in fp or "_P" in fp or "_W" in fp or "H2018" in fp or "P2018" in fp or "W2018" in fp


def read_raw(household_file, person_file, trip_file):
    """ Read the input files into format used by conversion """

    hh = pd.read_csv(household_file, encoding="windows-1252", delimiter=";", decimal=",",
                     quotechar="\"", low_memory=False, quoting=2)

    p = pd.read_csv(person_file, encoding="windows-1252", delimiter=";", decimal=",",
                    quotechar="\"", low_memory=False, quoting=2)

    t = pd.read_csv(trip_file, encoding="windows-1252", delimiter=";", decimal=",",
                    quotechar="\"", low_memory=False, quoting=2)

    return hh, p, t


def convert(data: tuple, regio=None):
    """ Convert srv data to standardized survey format """

    (hh, pp, tt) = data

    if regio is not None:
        regio = pd.read_csv(regio)

    ps = []
    for p in pp.itertuples():
        ps.append(
            Person(
                str(int(p.HHNR)) + "_" + str(int(p.PNR)),
                p.GEWICHT_P,
                str(int(p.HHNR)),
                int(p.V_ALTER),
                SrV2018.gender(p.V_GESCHLECHT),
                SrV2018.employment(p.V_ERW),
                False if p.V_EINSCHR_NEIN else True,
                SrV2018.yes_no(p.V_FUEHR_PKW),
                SrV2018.veh_avail(p.V_PKW_VERFUEG),
                Availability.YES if SrV2018.veh_avail(p.V_RAD_VERFUEG) == Availability.YES or SrV2018.veh_avail(
                    p.V_ERAD_VERFUEG) == Availability.YES else SrV2018.veh_avail(p.V_RAD_VERFUEG),
                SrV2018.veh_avail(p.V_FK_VERFUEG),
                p.V_WOHNUNG == 1,
                p.V_WOHNORT == 1,
                int(p.STICHTAG_WTAG),
                int(p.E_ANZ_WEGE)
            )
        )

    ps = pd.DataFrame(ps).set_index("p_id")

    random_state = np.random.RandomState(0)

    hhs = []
    for h in hh.itertuples():

        # Invalid entries in certain files
        if np.isnan(h.HHNR):
            continue

        hh_id = str(int(h.HHNR))
        hhs.append(
            Household(
                hh_id,
                h.GEWICHT_HH,
                pint(h.V_ANZ_PERS),
                pint(h.V_ANZ_PKW_PRIV + h.V_ANZ_PKW_DIENST),
                pint(h.V_ANZ_RAD + h.V_ANZ_ERAD),
                pint(h.V_ANZ_MOT125 + h.V_ANZ_MOPMOT + h.V_ANZ_SONST),
                SrV2018.parking_position(h.V_STELLPL1),
                SrV2018.economic_status(h.E_OEK_STATUS if "E_OEK_STATUS" in hh.keys() else -1, h.V_EINK,
                                        ps[ps.hh_id == hh_id]),
                SrV2018.household_type(h.E_HHTYP),
                SrV2018.region_type(h, regio, random_state),
                h.ST_CODE_NAME,
                zone=SrV2018.parse_zone(h),
                income=SrV2018.income(h.V_EINK),
            )
        )

    ts = []
    for t in tt.itertuples():
        # TODO: E_DAUER, E_GESCHW
        # E_ANKUNFT
        # TODO: double check
        ts.append(
            Trip(
                str(int(t.HHNR)) + "_" + str(int(t.PNR)) + "_" + str(int(t.WNR)),
                t.GEWICHT_W,
                str(int(t.HHNR)) + "_" + str(int(t.PNR)),
                str(int(t.HHNR)),
                int(t.WNR),
                int(t.STICHTAG_WTAG),
                int(t.E_BEGINN),
                int(t.E_DAUER),
                float(t.GIS_LAENGE),
                SrV2018.trip_mode(t.E_HVM),
                SrV2018.trip_purpose(t.V_ZWECK),
                SrV2018.sd_group(int(t.E_QZG_17)),
                # Trip is valid if length and duration are present
                0 <= t.GIS_LAENGE and t.E_DAUER > 0,
                from_location=SrV2018.parse_location(t, "V_START_"),
                from_zone=SrV2018.parse_zone(t, "V_START_"),
                to_location=SrV2018.parse_location(t, "V_ZIEL_"),
                to_zone=SrV2018.parse_zone(t, "V_ZIEL_"),
            )
        )

    return pd.DataFrame(hhs).set_index("hh_id"), ps, pd.DataFrame(ts).set_index("t_id")


def pint(x):
    """ Convert to positive integer"""
    return max(0, int(x))


def parse_int_str(x):
    """ Return parsed int or string """
    try:
        i = int(x)
        if i >= 0:
            return str(i)
        return None
    except ValueError:
        if not x:
            return None
        return x


class SrV2018:
    """ Maps SrV data to standard format"""

    # Fallback regiostar codes hard-coded
    CODES = {
        'Berlin': 1,
        'Potsdam': 2,
        'Cottbus': 4,
        'Frankfurt (Oder)': 2,
        'Kleinmachnow/Stahnsdorf/Teltow': 3,
        'Werder (Havel)/Schwielowsee': 3,
        'Michendorf/Nuthetal': 3,
        'Bernau': 3,
        'Blankenfelde-Mahlow/Rangsdorf': 3,
        'Dallgow/Falkensee/Wustermark': 3,
        'Eichwalde/Zeuthen': 3,
        'Fürstenwalde': 3,
        'Spremberg': 3,
        'Strausberg': 3,
        'Rüdersdorf': 3,
        'Oranienburg': 3,
        'Hennigsdorf/Velten': 3,
        'Brandenburg (Havel)': 5,
        'Schönefeld': 3,
        'Königs Wusterhausen': 3,
        'Eberswalde': 3,
        'Lübben/Lübbenau': 6,
        'Ludwigsfelde': 3,
        'Leipzig': 1
    }

    @staticmethod
    def parking_position(x):

        if x == 1:
            return ParkingPosition.PRIVATE
        elif x == 2:
            return ParkingPosition.PUBLIC
        elif x == 3:
            return ParkingPosition.DIFFERENT

        return ParkingPosition.NA

    @staticmethod
    def economic_status(status, eink, persons):

        if status >= 1:
            return list(EconomicStatus)[int(status) - 1]

        # Calculated according to Srv 2018
        # https://tu-dresden.de/bu/verkehr/ivs/srv/ressourcen/dateien/SrV2018_Tabellenbericht_Oberzentren_500TEW-_flach.pdf?lang=de

        if eink < 1:
            return EconomicStatus.UNKNOWN

        if eink == 1 or eink == 2:
            return EconomicStatus.VERY_LOW

        children = (persons.age < 14).sum()
        rest = len(persons) - children - 1

        w = 0.3 * children + 1 + 0.5 * rest

        if eink == 3:
            if w < 1.3:
                return EconomicStatus.LOW

            return EconomicStatus.VERY_LOW

        elif eink == 4:
            if w < 1.3:
                return EconomicStatus.MEDIUM
            elif w < 1.6:
                return EconomicStatus.LOW

            return EconomicStatus.VERY_LOW

        elif eink == 5:
            if w < 1.6:
                return EconomicStatus.MEDIUM
            elif w < 2.3:
                return EconomicStatus.LOW

            return EconomicStatus.VERY_LOW

        elif eink == 6:
            if w < 1.3:
                return EconomicStatus.HIGH
            elif w < 2.3:
                return EconomicStatus.MEDIUM
            elif w < 3.0:
                return EconomicStatus.LOW

            return EconomicStatus.VERY_LOW

        elif eink == 7:
            if w < 1.6:
                return EconomicStatus.HIGH
            elif w < 2.3:
                return EconomicStatus.MEDIUM
            elif w < 3.5:
                return EconomicStatus.LOW

            return EconomicStatus.VERY_LOW

        elif eink == 8:
            if w < 2.1:
                return EconomicStatus.HIGH
            elif w < 3.0:
                return EconomicStatus.MEDIUM

            return EconomicStatus.LOW

        elif eink == 9:
            if w < 1.3:
                return EconomicStatus.VERY_HIGH
            if w < 2.8:
                return EconomicStatus.HIGH
            return EconomicStatus.MEDIUM

        elif eink == 10:
            if w < 2.5:
                return EconomicStatus.VERY_HIGH

            return EconomicStatus.HIGH

        return EconomicStatus.UNKNOWN

    @staticmethod
    def household_type(x):
        if x == 1:
            return HouseholdType.MULTI_W_CHILDREN
        elif x == 2:
            return HouseholdType.MULTI_WO_CHILDREN

        return HouseholdType.SINGLE

    @staticmethod
    def gender(x):
        if x == 1:
            return Gender.M
        elif x == 2:
            return Gender.F

        return Gender.OTHER

    @staticmethod
    def employment(x):
        if x == 1:
            return Employment.CHILD
        elif x == 2:
            return Employment.HOMEMAKER
        elif x == 3:
            return Employment.RETIREE
        elif x == 5:
            return Employment.UNEMPLOYED
        elif x == 6:
            return Employment.SCHOOL
        elif x == 7:
            return Employment.STUDENT
        elif x == 8:
            return Employment.TRAINEE
        elif x == 9:
            return Employment.JOB_FULL_TIME
        elif x == 10 or x == 11:
            return Employment.JOB_PART_TIME

        return Employment.OTHER

    @staticmethod
    def yes_no(x):
        if x == 1:
            return Availability.YES
        elif x == 2:
            return Availability.NO

        return Availability.UNKNOWN

    @staticmethod
    def veh_avail(x):
        if x == 1 or x == 2:
            return Availability.YES
        elif x == 3:
            return Availability.NO

        return Availability.UNKNOWN

    @staticmethod
    def trip_mode(x):
        if x == 1:
            return TripMode.WALK
        elif x == 2 or x == 18 or x == 19:
            return TripMode.BIKE
        elif x == 3:
            return TripMode.MOTORCYCLE
        elif x == 4 or x == 5 or x == 6:
            return TripMode.CAR
        elif x == 7 or x == 8 or x == 9:
            return TripMode.RIDE
        elif 10 <= x <= 17:
            return TripMode.PT

        return TripMode.OTHER

    @staticmethod
    def trip_purpose(x):
        if x == 1:
            return Purpose.WORK
        elif x == 2:
            return Purpose.WORK_BUSINESS
        elif x == 3:
            return Purpose.EDU_KIGA
        elif x == 4:
            return Purpose.EDU_PRIMARY
        elif x == 5:
            return Purpose.EDU_SECONDARY
        elif x == 6:
            return Purpose.EDU_HIGHER
        elif x == 7:
            return Purpose.EDU_OTHER
        elif x == 8:
            return Purpose.SHOP_DAILY
        elif x == 9:
            return Purpose.SHOP_OTHER
        elif x == 19 or x == 10:
            return Purpose.PERSONAL_BUSINESS
        elif x == 11:
            return Purpose.TRANSPORT
        elif x == 15:
            return Purpose.OUTSIDE_RECREATION
        elif x == 13:
            return Purpose.DINING
        elif 12 <= x <= 17:
            return Purpose.LEISURE
        elif x == 18:
            return Purpose.HOME

        return Purpose.OTHER

    @staticmethod
    def sd_group(x):
        if x <= 0:
            return SourceDestinationGroup.UNKNOWN

        return list(SourceDestinationGroup)[int(x) - 1]

    @staticmethod
    def region_type(d, regio, random_state):

        if "PLZ" in dir(d) and regio is not None:
            plz = int(d.PLZ)

            r = regio[regio.plz == plz]
            if len(r) == 0:
                raise ValueError("Unknown PLZ: %d" % plz)

            if len(r) > 1:
                r = r.sample(n=1, random_state=random_state)

            r = r.iloc[0].RegioStaR7
            if np.isnan(r):
                raise ValueError("No RegioStar value available for: %d" % plz)

            return int(r) - 70

        else:
            name = d.ST_CODE_NAME
            return SrV2018.CODES[name]

    @staticmethod
    def income(x):
        if x == 1:
            return 0
        elif x == 2:
            return 500
        elif x == 3:
            return 900
        elif x == 4:
            return 1500
        elif x == 5:
            return 2000
        elif x == 6:
            return 2600
        elif x == 7:
            return 3000
        elif x == 8:
            return 3600
        elif x == 9:
            return 4600
        elif x == 10:
            return 5600

        return -1

    @staticmethod
    def parse_zone(h, prefix=""):
        ob = parse_int_str(getattr(h, prefix + "OBERBEZIRK"))

        zone = pd.NA
        if ob:
            zone = ob
            ub = parse_int_str(getattr(h, prefix + "UNTERBEZIRK"))
            if ub:
                zone += "-" + ub
                tb = parse_int_str(getattr(h, prefix + "TEILBEZIRK"))
                if tb:
                    zone += "-" + tb

        return zone

    @staticmethod
    def parse_location(h, prefix=""):
        ort = parse_int_str(getattr(h, prefix + "ORT"))
        if not ort:
            return pd.NA

        if "Berlin" in ort:
            return "Berlin"

        return ort



import xopen
import xml.etree.ElementTree as ET
import pandas as pd
from matsim import utils

class Household:
    def __init__(self, households):
        self.households = households

# Returns households as a dataframe
def houshold_reader(filename, convert_dataframes_types=True):
    tree = ET.iterparse(xopen.xopen(filename, 'r'), events=['start','end'])
    
    households = []
    current_persons = []
    current_household = {}
    
    for xml_event, elem in tree:
        _, _, elem_tag = elem.tag.partition('}')     # Removing xmlns tag from tag name
        
        # HOUSEHOLDS
        if elem_tag == 'household':
            if xml_event == 'start':
                utils.parse_attributes(elem, current_household)
            else:
                current_household['members'] = current_persons
                households.append(current_household)
                current_household = {}
                current_persons = []
                elem.clear()

        # ATTRIBUTES
        elif elem_tag == 'attribute':
            current_household[elem.attrib['name']] = elem.text
        
        # MEMBERS
        elif elem_tag == 'personId' and xml_event == 'start':
            current_persons.append(int(elem.attrib['refId']))
    
    
    # Convert to dataframe and converts columns types
    households = pd.DataFrame.from_records(households)
    if convert_dataframes_types:
        try:
            households['id'] = households['id'].astype(int)
            households['censusId'] = households['censusId'].astype(int)
            households['household_income'] = households['household_income'].astype(float)
        except KeyError:
            print('dataframe types conversion failed')
    
    return Household(households)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

import pandas as pd

METADATA = "data-plz-regiostar", "Create regiostar lookup for plz."


def prepare_regiostar_lookup(plz, regiostar) -> pd.DataFrame:
    plz = pd.read_csv(plz)

    regiostar = pd.read_excel(regiostar, sheet_name="ReferenzGebietsstand2020")
    df = plz.merge(regiostar, how="left", left_on="ags", right_on="gem_20")

    df = df[["plz", "ags", "RegioStaR2", "RegioStaR4", "RegioStaR5", "RegioStaR7", "RegioStaR17"]]

    return df


def setup(parser: argparse.ArgumentParser):
    pass


def main(args):
    d = "/Users/rakow/Development/matsim-scenarios/shared-svn/projects/matsim-germany/"

    # FIXME: use args

    df = prepare_regiostar_lookup(d + "zuordnung_plz_ort.csv", d + "RegioStaR-Referenzdateien.xlsx")

    df.to_csv(d + "zuordnung_plz_regiostar.csv", index=False)



# -*- coding: utf-8 -*-

from google.protobuf.internal.encoder import _EncodeVarint

from xopen import xopen

from .pb.Wireformat_pb2 import ContentType, PBFileHeader
from .pb.Events_pb2 import EventBatch

PB_VERSION = {
    ContentType.EVENTS: (1, EventBatch)
}

# Parses attributes of an element and adds them to the given dictionary
def parse_attributes(elem, dict):
    for attrib in elem.attrib:
        dict[attrib] = elem.attrib[attrib]
    return dict


def read_pb(filepath):
    """ Read MATSim protobuf file and yield each element """
    with xopen(filepath, "rb") as f:
        header = PBFileHeader()
        offset, pos = _read_varint(f)
        header.ParseFromString(f.read(offset))

        supported, msg = PB_VERSION[header.contentType]
        if supported < header.version:
            raise Exception("Unsupported protobuf version: %d" % header.version)

        for batch in read_delimited(msg, f):
            for ev in batch.events:
                yield ev


def write_delimited(msgs, write):
    """ Helper function to write delimited protobuf messages """
    encoder = _EncodeVarint
    for msg in msgs:
        buf = msg.SerializeToString()
        encoder(write, len(buf))
        write(buf)


def read_delimited(msg_class, stream):
    """ Helper function to read delimited protobuf messages from a buffer """
    while 1:
        m = msg_class()
        length, read = _read_varint(stream)
        if read == 0:
            return
        m.ParseFromString(stream.read(length))
        yield m


def _read_varint(stream):
    """
    read a varint from a stream
    :returns (result, bytes read)
    """
    # Mask for 32bit integer
    mask = (1 << 32) - 1

    result = 0
    shift = 0
    pos = 0
    while 1:
        buf = stream.read(1)
        if not buf:
            return 0, 0

        b = buf[0]
        result |= ((b & 0x7f) << shift)
        pos += 1
        if not (b & 0x80):
            result &= mask
            result = int(result)
            return result, pos
        shift += 7
        if shift >= 64:
            raise RuntimeError('Too many bytes when decoding varint.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path

import pandas as pd


def build_datasets(network, inter, routes, model_type):
    """ Build all datasets needed for training models"""
    ft = pd.concat([pd.read_csv(n) for n in network])

    df_i = pd.concat([pd.merge(pd.read_csv(i), ft, left_on="fromEdgeId", right_on="linkId") for i in inter])
    df_r = pd.concat(
        [pd.merge(pd.read_csv(r).drop(columns=["speed"]), ft, left_on="edgeId", right_on="linkId") for r in routes])

    result = {}

    aggr = df_r.groupby(["junction_type"])
    for g in aggr.groups:
        if str(g) == "dead_end":
            continue

        result["speedRelative_" + str(g)] = prepare_dataframe(aggr.get_group(g), model_type, target="speedRelative")

    aggr = df_i.groupby(["junction_type"])
    df_i["norm_cap"] = df_i.capacity / df_i.num_lanes
    for g in aggr.groups:
        result["capacity_" + str(g)] = prepare_dataframe(aggr.get_group(g), model_type, target="norm_cap")

    return result


def prepare_dataframe(df, model_type, target):
    """ Simple preprocessing """

    df = df.rename(columns={target: "target"})

    # Drop length outliers
    df = df[df.length < 500]

    # drop 2.5% smallest and largest
    drop = len(df) // 40
    df = df.drop(df.nsmallest(drop, ["target"]).index)
    df = df.drop(df.nlargest(drop, ["target"]).index)

    s = df['target']
    q1 = s.quantile(0.25)
    q3 = s.quantile(0.75)
    iqr = q3 - q1
    iqr_lower = q1 - 1.5 * iqr
    iqr_upper = q3 + 1.5 * iqr
    outliers = s[(s < iqr_lower) | (s > iqr_upper)]

    df = df.drop(outliers.index)

    # select features based on model
    if model_type == "extended":
        df = df[["target", "length", "speed", "change_speed", "num_lanes", "change_num_lanes", "num_to_links",
                 "dir_l", "dir_r", "dir_s", "dir_multiple_s", "dir_exclusive", "num_foes",
                 "junction_inc_lanes", "priority_lower", "priority_equal", "priority_higher",
                 "is_secondary_or_higher", "is_primary_or_higher", "is_motorway", "is_link"]]
    elif model_type == "default":
        df = df[["target", "length", "speed", "change_speed", "num_lanes", "change_num_lanes", "num_to_links",
                 "junction_inc_lanes", "priority_lower", "priority_equal", "priority_higher",
                 "is_secondary_or_higher", "is_primary_or_higher", "is_motorway", "is_link"]]

    else:
        raise ValueError("Illegal model type:" + model_type)

    return df


def read_edges(folder):
    """ Combine resulting files for edges """

    data = []
    for f in os.listdir(folder):
        if not f.endswith(".csv"):
            continue

        df = pd.read_csv(os.path.join(folder, f))
        edge_id = df.iloc[0].edgeId

        aggr = df.groupby("laneId").agg(capacity=("flow", "max"))

        data.append({
            "edgeId": edge_id,
            "capacity": float(aggr.capacity.mean())
        })

    return pd.DataFrame(data)


def read_intersections(folder):
    """ Read intersection results """

    data = []
    for f in os.listdir(folder):
        if not f.endswith(".csv"):
            continue

        try:
            df = pd.read_csv(os.path.join(folder, f))
        except pd.errors.EmptyDataError:
            print("Empty csv", f)
            continue

        # there could be exclusive lanes and the capacity to two edges completely additive
        # however if the capacity is shared one direction could use way more than physical possible
        aggr = df.groupby("fromEdgeId").agg(capacity=("flow", "max")).reset_index()
        aggr.rename(columns={"fromEdgeId": "edgeId"})

        data.append(aggr)

    return pd.concat(data)


def read_routes(folder):
    """ Read routes from folder """
    data = []
    for f in os.listdir(folder):
        if not f.endswith(".csv"):
            continue

        try:
            df = pd.read_csv(os.path.join(folder, f))
        except pd.errors.EmptyDataError:
            print("Empty csv", f)
            continue

        data.append(df)

    df = pd.concat(data)
    aggr = df.groupby("edgeId").mean()

    return aggr.reset_index()



#!/usr/bin/env python
# @author  Angelo Banse, Ronald Nippold, Christian Rakow

import os
import shutil
import sys
from argparse import ArgumentParser
from os.path import join, basename

from .utils import init_workload, setup_parser, create_args, write_scenario, filter_network, vehicle_parameter

import sumolib.net
import traci  # noqa
from sumolib import checkBinary  # noqa
import lxml.etree as ET

import pandas as pd
import numpy as np

sumoBinary = checkBinary('sumo')
netconvert = checkBinary('netconvert')

METADATA = "sumo-edges", "Determine edge volumes with SUMO."


def capacity_estimate(v):
    tT = 1.2
    lL = 7.0
    Qc = v / (v * tT + lL)

    return 3600 * Qc


def writeRouteFile(f_name, departLane, arrivalLane, edges, veh, scenario):
    text = """<?xml version="1.0" encoding="UTF-8"?>
<routes xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/routes_file.xsd">

"""
    text += """<vTypeDistribution id="vDist">
                    %s
                </vTypeDistribution>
            """ % vehicle_parameter(scenario)

    text += """
    <flow id="veh" begin="0" end= "600" vehsPerHour="{veh}" type="vDist" departLane="{departLane}" arrivalLane="{arrivalLane}" departSpeed="max">
        <route edges="{edges}"/>
    </flow>

</routes>"""

    # departSpeed="speedLimit" ?
    context = {
        "departLane": departLane,
        "arrivalLane": arrivalLane,
        "edges": edges,
        "veh": veh
    }

    with open(f_name, "w") as f:
        f.write(text.format(**context))


def writeDetectorFile(f_name, output, lane, laneNr, scale):
    text = """<?xml version="1.0" encoding="UTF-8"?>

	<additional xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/additional_file.xsd">
	        %s
	</additional>

	""" % "\n".join(
        """<e1Detector id="detector_%d" lane="{lane}_%d" pos="-15" friendlyPos="true" freq="10.00" file="{output_file}_%d.xml"/>""" % (
            i, i, i)
        for i in
        range(laneNr))

    context = {
        "lane": lane,
        "laneNr": laneNr,
        "output_file": join(output, scale, "lane")
    }

    with open(f_name, 'w') as f:
        f.write(text.format(**context))


def read_result(folder, edge, scale):
    data = []

    for f in os.listdir(folder):
        if not f.endswith(".xml"):
            continue

        total = 0
        end = 0

        for _, elem in ET.iterparse(join(folder, f), events=("end",),
                                    tag=('interval',),
                                    remove_blank_text=True):

            begin = float(elem.attrib["begin"])
            end = float(elem.attrib["end"])
            if begin < 60:
                continue

            total += float(elem.attrib["nVehContrib"])

        data.append({
            "edgeId": edge,
            "laneId": f.replace(".xml", ""),
            "flow": total * (3600 / (end - 60)),
            "scale": float(scale),
            "count": total
        })

    return data


def run(args, edges):
    # saveToFile(edges_ids,"junctions.json")
    i = 0

    if args.to_index <= 0:
        args.to_index = len(edges)

    for x in range(args.from_index, args.to_index):
        edge = edges[x]
        i += 1
        print("Edge id: ", edge._id)
        print("Number of lanes: ", edge.getLaneNumber(), "speed:", edge.getSpeed())

        laneNr = edge.getLaneNumber()  # nr of lanes

        cap = capacity_estimate(edge.getSpeed()) * 0.9 * laneNr

        print("Capacity estimate:", cap)

        p_network = join(args.runner, "filtered.net.xml")
        p_routes = join(args.runner, "route.rou.xml")
        p_detector = join(args.runner, "detector.add.xml")

        filter_network(netconvert, args.network, edge, p_network)
        writeRouteFile(p_routes, "best", "current", edge._id, cap, args.scenario)
        p_scenario = join(args.runner, "scenario.sumocfg")

        write_scenario(p_scenario, basename(p_network), basename(p_routes), basename(p_detector), args.step_length)

        go(p_scenario, p_network, edge, p_detector, args)
        print("####################################################################")
        print("[" + str(i) + " / " + str(args.to_index - args.from_index) + "]")


def go(scenario, network, edge, p_detector, args):
    # while traci.simulation.getMinExpectedNumber() > 0:

    end = int(600 * (1 / args.step_length))

    res = []

    folder = join(args.runner, "detector")

    # Clean old data
    shutil.rmtree(folder, ignore_errors=True)
    os.makedirs(folder, exist_ok=True)

    traci.start([sumoBinary, "-n", network], port=args.port)

    xr = ["%.2f" % s for s in np.arange(1, 2.1, 0.05)]

    # Simulate different scales
    for scale in xr:

        # print("Running scale", scale)

        os.makedirs(join(folder, scale), exist_ok=True)
        writeDetectorFile(p_detector, "detector", edge._id, edge.getLaneNumber(), scale)

        # Load scenario with desired traffic scaling
        traci.load(["-c", scenario, "--scale", scale])

        try:
            for step in range(0, end):
                traci.simulationStep()
        except Exception as e:
            print(e)

    traci.close()

    for scale in xr:
        res.extend(read_result(join(folder, scale), edge._id, scale))

    df = pd.DataFrame(res)
    df.to_csv(join(args.output, "%s.csv" % edge._id), index=False)

    sys.stdout.flush()


def setup(parser: ArgumentParser):
    setup_parser(parser)


def main(args):
    args = create_args(args)

    net = sumolib.net.readNet(args.network, withConnections=False, withInternal=False, withFoes=False)

    allEdges = net.getEdges()  # all type of edges

    with open(args.input[0]) as f:
        selection = set(f.read().splitlines())

    # select if edges in net file
    edges = [edge for edge in allEdges if edge._id in selection]

    init_workload(args, edges)

    print("Total number of edges:", len(edges))
    print("Processing: ", args.from_index, ' to ', args.to_index)

    run(args, edges)


if __name__ == "__main__":
    parser = ArgumentParser(prog=METADATA[0], description=METADATA[1])
    setup(parser)
    main(parser.parse_args())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Sequence, SupportsFloat

import numpy as np
import pandas as pd

from . import *
from .preparation import cut

METADATA = "data-create-ref", "Extract and create reference data from surveys."


class InvalidHandling(Enum):
    """ How to handle invalid trips. """
    # Invalid trips are ignored
    KEEP = auto()
    # Drop single invalid trips
    REMOVE_TRIPS = auto()
    # Drop whole person if any trip is invalid
    REMOVE_PERSONS = auto()


@dataclass
class AggregationResult:
    """ Return value of create function. """

    persons: pd.DataFrame
    trips: pd.DataFrame
    share: pd.DataFrame

    groups: pd.DataFrame = None


def weighted(x):
    data = dict(n=x.t_weight.sum())
    return pd.Series(data=data)


def summarize_purposes(x):
    x["departure_h"] = x.departure // 60
    x["arrival_h"] = (x.departure + x.duration) // 60

    d = x.groupby(["purpose", "departure_h"]).agg(n=("t_weight", "sum"))
    d["departure"] = d.n / d.n.sum()
    d = d.drop(columns=["n"])
    d.index.rename(names=["purpose", "h"], inplace=True)

    a = x.groupby(["purpose", "arrival_h"]).agg(n=("t_weight", "sum"))
    a["arrival"] = a.n / a.n.sum()
    a = a.drop(columns=["n"]).rename(index={"arrival_h": "h"})
    a.index.rename(names=["purpose", "h"], inplace=True)

    m = pd.merge(a, d, left_index=True, right_index=True, how="outer")
    m.fillna(0, inplace=True)

    return m


def mode_usage(mode):
    def f(x):
        return (x == mode).any()

    return f


def grouped_share(df, groups, normalize=True):
    """ Compute share of each group. """

    aggr = df.groupby(groups).apply(weighted)
    aggr["share"] = aggr.n / aggr.n.sum()
    aggr["share"].fillna(0, inplace=True)
    aggr.drop(columns=["n"], inplace=True)

    # Normalize per group
    if normalize:
        for group in set(aggr.index.get_level_values(0)):
            sub = aggr.loc[group, :]
            sub.share /= sub.share.sum()

    return aggr


def summarize_mode_usage(x, trips):
    total_mobile = x[x.mobile_on_day].p_weight.sum()

    args = {k.value: ("main_mode", mode_usage(k)) for k in set(trips.main_mode)}

    p_trips = trips[trips.p_id.isin(x.index)]

    mode_user = p_trips.groupby(["p_id"]).agg(**args)
    joined = x.join(mode_user, how="inner")

    data = {}
    for c in mode_user.columns:
        share = joined[joined[c]].p_weight.sum() / total_mobile
        data[c] = share

    return pd.DataFrame(data={"main_mode": data.keys(), "user": data.values()}).set_index("main_mode")


def setup(parser: argparse.ArgumentParser):
    parser.add_argument("dirs", nargs="+", help="Directories with survey data")


def default_person_filter(df):
    """ Default person filter for reference data. """
    return df[df.present_on_day & (df.reporting_day <= 4)]


def create(survey_dirs, transform_persons, transform_trips,
           invalid_trip_handling: InvalidHandling = InvalidHandling.REMOVE_TRIPS,
           dist_groups: Sequence[SupportsFloat] = None,
           ref_groups: List[str] = None,
           output_prefix="") -> AggregationResult:
    """ Create reference data from survey data.
    :param survey_dirs: Directories with survey data
    :param transform_persons: Function to transform person data frame
    :param transform_trips: Function to transform trip data frame
    :param invalid_trip_handling: How to handle invalid trips
    :param dist_groups: distance group bins in meters
    :param ref_groups: Create reference data for these attribute groups
    :param output_prefix: prefix for the ouput files
    :return:
    """

    if dist_groups is None:
        dist_groups = [0, 1000, 2000, 5000, 10000, 20000, np.inf]

    all_hh, all_persons, all_trips = read_all(survey_dirs)

    # Filter person ad trips for area
    df = all_persons.join(all_hh, on="hh_id")

    persons = transform_persons(df) if transform_persons is not None else df

    if invalid_trip_handling == InvalidHandling.REMOVE_PERSONS:
        # Filter persons, if they have at least one invalid trip
        invalid = set(all_trips[~all_trips.valid].p_id)
        persons = persons[~persons.index.isin(invalid)]
    elif invalid_trip_handling == InvalidHandling.REMOVE_TRIPS:
        # Use only valid trips
        all_trips = all_trips[all_trips.valid]

    # Because of inner join, trips might be filtered if person was removed
    trips = all_trips.drop(columns=["hh_id"]).join(persons, on="p_id", how="inner")

    # Transform existing trips
    trips = transform_trips(trips) if transform_trips is not None else trips

    # Set dist groups
    trips["dist_group"] = cut(trips.gis_length * 1000, dist_groups)

    aggr = trips.groupby(["dist_group", "main_mode"]).apply(weighted)

    aggr["share"] = aggr.n / aggr.n.sum()
    aggr["share"].fillna(0, inplace=True)

    share = aggr.drop(columns=["n"])
    aggr = share.copy()

    aggr.to_csv(output_prefix + "mode_share_ref.csv")

    # Also normalize per distance group
    for dist_group in aggr.index.get_level_values(0).categories:
        sub = aggr.loc[dist_group, :]
        sub.share /= sub.share.sum()

    aggr.to_csv(output_prefix + "mode_share_per_dist_ref.csv")

    aggr = summarize_purposes(trips)

    aggr.to_csv(output_prefix + "trip_purposes_by_hour_ref.csv")

    aggr = summarize_mode_usage(persons, trips)
    aggr.to_csv(output_prefix + "mode_users_ref.csv")

    groups = None
    if ref_groups:

        overall = share.groupby("main_mode").sum().reset_index()

        # Include total share in reference data
        groups = [overall]
        dist = [grouped_share(trips, ["dist_group", "main_mode"], normalize=False).reset_index()]

        for g in ref_groups:

            if g not in persons.columns:
                raise ValueError("Column %s not found in persons" % g)

            aggr = grouped_share(trips, [g, "main_mode"])
            groups.append(aggr.reset_index())

            aggr = grouped_share(trips, [g, "dist_group", "main_mode"])
            dist.append(aggr.reset_index())

        groups = pd.concat(groups, sort=False)
        # Reorder columns
        groups = groups[ref_groups + ["main_mode", "share"]]
        groups.to_csv(output_prefix + "mode_share_per_group_ref.csv", index=False)

        dist = pd.concat(dist, sort=False)
        dist = dist[ref_groups + ["dist_group", "main_mode", "share"]]
        dist.to_csv(output_prefix + "mode_share_per_group_dist_ref.csv", index=False)

    return AggregationResult(persons, trips, share.groupby("main_mode").sum(), groups=groups)


def main(args):
    create(args.dirs, default_person_filter)



# -*- coding: utf-8 -*-

# TODO: collection of old code
# was just draft anyway, needs to be updated

"""
m["marginalUtilityOfDistance_util_m"] = trial.suggest_float("dist_" + mode, sys.float_info.min,
                                                                            sys.float_info.max)


def sample_dist_util(self, mode, last):

        df = pd.DataFrame.from_dict(last.user_attrs["mode_stats"], orient="tight")
        target = self.target[self.target["mode"] == mode].reset_index(drop=True).copy()

        df = df.loc[target.dist_group].reset_index()

        # Trips distances shares over all modes
        # Correction factor is calculated here
        ref_dist = self.target.groupby("dist_group").agg(share=("share", "sum"))
        sim_dist = df.groupby("dist_group").agg(share=("share", "sum"))

        correction = ref_dist.loc[target.dist_group] / sim_dist.loc[target.dist_group]

        df = df[df.main_mode == mode].reset_index(drop=True).copy()

        if "mean_dist" not in target.columns:
            target["mean_dist"] = df.mean_dist

        df["correction"] = correction.values

        target.share = target.share / sum(target.share)
        df.share = df.share / sum(df.share)

        real = (df.mean_dist * df.share * df.correction).sum()
        target = (target.mean_dist * target.share).sum()

        # TODO: magnitude should depend on the asc

        # TODO: configurable parameter
        return float(0.05 * (target - real) / (1000 * 1000))

"""


# -*- coding: utf-8 -*-
"""These models depend on the feyn package which is not listed and only optional due to its licensing"""

from time import time

import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, RegressorMixin, ClassifierMixin
from sklearn.metrics import mean_squared_error

import feyn

class QLatticeBase(BaseEstimator):
    _kind = None
    _loss = None

    @staticmethod
    def as_df(X, y=None):
        tf = pd.DataFrame(X, dtype=np.float64)
        if y is not None:
            tf["y"] = y

        tf.columns = tf.columns.map(lambda x: "data%s" % x if x != "y" else "y")

        return tf

    def __init__(self, max_complexity, n_epochs, criterion, progress, random_state) -> None:
        super().__init__()

        self.ql = feyn.QLattice(random_state)
        self.n_epochs = n_epochs
        self.max_complexity = max_complexity
        self.criterion = criterion
        self.progress = progress
        self.random_state = random_state

    def fit(self, X, y, sample_weight=None, test=None, val=None, f=None, ftest=None, fval=None, val_interval=10):

        tf = self.as_df(X, y)

        if test is not None:
            test = self.as_df(*test)

        if val is not None:
            val = self.as_df(*val)

        models = []
        m_count = 0

        start = time()
        for i in range(1, self.n_epochs + 1):

            new_sample = self.ql.sample_models(
                input_names=tf.columns,
                output_name='y',
                kind=self._kind,
                max_complexity=self.max_complexity
            )

            models += new_sample
            m_count += len(new_sample)

            models = feyn.fit_models(
                models=models,
                data=tf,
                sample_weights=sample_weight,
                loss_function=self._loss,
                criterion=self.criterion,
                threads=8
            )

            # Sort for val method for pruning
            if test is not None:
                if f is None:
                    f = lambda pred: mean_squared_error(tf["y"], pred)

                if ftest is None:
                    ftest = lambda pred: mean_squared_error(test["y"], pred)

                if fval is None:
                    fval = lambda pred: mean_squared_error(val["y"], pred)

                # Custom prune / sorting
                if val is not None and ((i - 1) % val_interval == 0 or i == self.n_epochs):
                    models = list(sorted(
                        models, key=lambda m: ftest(m.predict(test)) + fval(m.predict(val))
                    ))
                else:
                    models = list(sorted(
                        models, key=lambda m: f(m.predict(tf)) + ftest(m.predict(test))
                    ))

            models = feyn.prune_models(models)

            info = ""
            if ftest is not None:
                # Custom label
                info += " Test: %.2f" % ftest(models[0].predict(test))

            if val is not None and fval is not None:
                # Custom label
                info += " Val: %.2f" % fval(models[0].predict(val))

            elapsed = time() - start
            if len(models) > 0 and self.progress:
                feyn.show_model(
                    models[0],
                    feyn.tools.get_progress_label(i, self.n_epochs, elapsed, m_count) + info,
                    update_display=True,
                    )

            self.ql.update(models)

        self.best = [models[0]] + feyn.get_diverse_models(models[1:], n=5)

        return self

    def predict(self, X):
        return self.best[0].predict(self.as_df(X))

    def show(self):
        feyn.show_model(self.best[0])

    def sympify(self):
        return self.best[0].sympify(signif=9, symbolic_lr=True, include_weights=True)

    def plot(self, X, y, Xval, yval):

        train = self.as_df(X, y)
        test = self.as_df(Xval, yval)

        self.best[0].plot(
            data=train,
            compare_data=test
        )


class QLatticeClassifier(QLatticeBase, ClassifierMixin):

    def __init__(self, max_complexity=10, n_epochs=30, criterion="bic", progress=False, random_state=-1) -> None:
        super().__init__(max_complexity, n_epochs, criterion, progress, random_state)

        self._kind = "classification"
        self._loss = "binary_cross_entropy"

    def copy(self, n=0):
        """ Make new model by copying the best n"""

        other = QLatticeClassifier(self.max_complexity, self.n_epochs, self.progress, self.random_state)
        other.best = [self.best[n]]

        return other


class QLatticeRegressor(QLatticeBase, RegressorMixin):

    def __init__(self, max_complexity=10, n_epochs=30, criterion="bic", progress=False, random_state=-1) -> None:
        super().__init__(max_complexity, n_epochs, criterion, progress, random_state)

        self._kind = "regression"
        self._loss = "squared_error"

    def copy(self, n=0):
        """ Make new model by copying the best n"""

        other = QLatticeRegressor(self.max_complexity, self.n_epochs, self.progress, self.random_state)
        other.best = [self.best[n]]

        return other


# -*- coding: utf-8 -*-

from pathlib import Path
from collections.abc import Iterable
import xml.etree.ElementTree as ET
import json

from xopen import xopen

from .pb.Ids_pb2 import ProtoId
from .pb.Model_pb2 import Coordinate

from matsim.utils import read_pb


def event_reader(filepath, types=None):
    """ Reads an events file in any of the supported formats (xml, json, pb) and yields each contained event.
        Each event will be generated as a dictionary of attribute key/value pairs.

    :param filepath path to the file
    :param types event types to return. Can be an iterable, or comma-separated string. Default None returns all events.
    :returns generator of events from the specified file
    :rtype Iterable[dict]
    """
    # set up event filter - so that we only yield useful events
    if types is None:
        keep = None
    elif isinstance(types, str):
        keep = set(types.split(','))
    elif isinstance(types, Iterable):
        keep = set(types)
    else:
        raise ValueError("Invalid argument for types: %s" % type(types))

    if isinstance(filepath, Path):
        filepath = str(filepath)

    if '.xml' in filepath:
        reader = _event_reader_xml
    elif '.pb' in filepath:
        reader = _event_reader_pb
    elif '.ndjson' in filepath:
        reader = _event_reader_json
    else:
        raise ValueError('Format of %s unknown or not supported' % filepath)

    for event in reader(filepath):
        # skip events we don't care about
        if keep and not event['type'] in keep:
            continue

        yield event


def _event_reader_xml(filepath):
    """ Any content text of the XML element itself is dropped, as MATSim events are attribute-only. """
    with xopen(filepath) as f:
        tree = ET.iterparse(f, events=['start', 'end'])
        _, root = next(tree)
        try:
            for xml_event, elem in tree:
                attributes = elem.attrib

                if xml_event == 'start' and elem.tag == 'event':
                    # got one! yield the event to the caller
                    attributes['time'] = float(attributes['time'])
                    yield attributes

                # free memory. Otherwise the data is kept in memory and we loose the advantage of streaming
                root.clear()

        except Exception as e:
            # Why am I catching this exception instead of allowing it to propagate up?
            # Because some event files (**coughswitzerland**) are badly formed and do not contain
            # the closing </events> tag at the end of the file. If we don't trap that error, the
            # entire analysis fails.
            print('*** XML ERROR:', e)


def _event_reader_pb(filepath, convert=True):
    for event in read_pb(filepath):
        # the conversion nullifies protobufs performance advantage
        if convert:
            yield _convert_pb_event(event)
        else:
            yield event

# Mapping of protobuf types to names
MAPPING = {'activityEnd': 'actend', 'personDeparture': 'departure', 'personEntersVehicle': 'PersonEntersVehicle',
           'vehicleEntersTraffic': 'vehicle enters traffic', 'linkLeave': 'left link', 'linkEnter': 'entered link',
           'vehicleLeavesTraffic': 'vehicle leaves traffic', 'personLeavesVehicle': 'PersonLeavesVehicle',
           'personalArrival': 'arrival', 'activityStart': 'actstart'}


def _convert_pb_event(event):
    entry = {}
    ev_type = event.WhichOneof("type")
    entry["type"] = MAPPING.get(ev_type, ev_type)
    for (desc, v) in event.ListFields():
        if ev_type == desc.name:
            for (attr, evValue) in v.ListFields():
                if type(evValue) == ProtoId:
                    entry[attr.name] = evValue.id

        elif type(v) == Coordinate:
            entry["x"] = v.x
            entry["y"] = v.y
        else:
            entry[desc.name] = v

    return entry


def _event_reader_json(filepath):
    with xopen(filepath) as f:
        for line in f:
            yield json.loads(line)



#!/usr/bin/env python
# @author  Angelo Banse, Ronald Nippold, Christian Rakow

import os
import shutil
import sys
from argparse import ArgumentParser
from os.path import join, basename

from .utils import setup_parser, create_args, init_workload, write_scenario, filter_network, vehicle_parameter

import traci  # noqa
import sumolib.net
from sumolib import checkBinary  # noqa
import lxml.etree as ET

import pandas as pd

sumoBinary = checkBinary('sumo')
netconvert = checkBinary('netconvert')

METADATA = "sumo-intersections", "Determine intersection volumes with SUMO."


def writeRouteFile(f_name, routes, extra_routes, scenario):
    """ Write route file for intersection """

    text = """<?xml version="1.0" encoding="UTF-8"?>

<routes xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/routes_file.xsd">


"""
    text += """<vTypeDistribution id="vDist">
                    %s
                </vTypeDistribution>
            """ % vehicle_parameter(scenario)

    for i, edges in enumerate(routes):
        text += """
            <flow id="veh%d" begin="0" end="1800" vehsPerHour="5000" type="vDist" departLane="best" arrivalLane="current" departSpeed="max">
               <route edges="%s"/>
            </flow>
        """ % (i, edges)

    for i, edges in enumerate(extra_routes):
        text += """
            <flow id="vehx%d" begin="0" end="1800" vehsPerHour="200" type="vDist" departLane="best" arrivalLane="current" departSpeed="max">
               <route edges="%s"/>
            </flow>
        """ % (i, edges)

    text += "</routes>"

    with open(f_name, "w") as f:
        f.write(text)


def writeDetectorFile(f_name, output, lanes):
    text = """<?xml version="1.0" encoding="UTF-8"?>
        <additional xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/additional_file.xsd">
"""

    for i, lane in enumerate(lanes):
        text += """
        <e1Detector id="detector%d" lane="%s" pos="-1" friendlyPos="true" freq="10.00" file="%s.xml"/>
    """ % (i, lane, join(output, "lane%d" % i))

    text += "</additional>"

    with open(f_name, 'w') as f:
        f.write(text)


def read_result(folder, **kwargs):
    flow = 0

    for f in os.listdir(folder):
        if not f.endswith(".xml"):
            continue

        total = 0
        end = 0

        for _, elem in ET.iterparse(join(folder, f), events=("end",),
                                    tag=('interval',),
                                    remove_blank_text=True):

            begin = float(elem.attrib["begin"])
            end = float(elem.attrib["end"])
            if begin < 60:
                continue

            total += float(elem.attrib["nVehContrib"])

        flow += total * (3600 / (end - 60))

    kwargs["flow"] = flow
    return kwargs


def run(args, nodes):
    print("Running scenario: " + args.scenario)

    if args.to_index <= 0:
        args.to_index = len(nodes)

    i = 0

    for x in range(args.from_index, args.to_index):
        node = nodes[x]
        i += 1

        print("####################################################################")
        print("Junction id: " + node._id)

        folder = join(args.runner, "detector")
        p_network = join(args.runner, "filtered.net.xml")

        edges = [c.getFrom() for c in node.getConnections()] + [c.getTo() for c in node.getConnections()]

        filter_network(netconvert, args.network, edges, p_network, ["--no-internal-links", "false"])

        pairs = set((c.getFrom(), c.getTo()) for c in node.getConnections() if c._direction != c.LINKDIR_TURN)

        res = []

        for fromEdge, toEdge in pairs:

            # Clean old data
            shutil.rmtree(folder, ignore_errors=True)
            os.makedirs(folder, exist_ok=True)

            p_scenario = join(args.runner, "scenario.sumocfg")
            p_routes = join(args.runner, "route.rou.xml")
            p_detector = join(args.runner, "detector.add.xml")

            routes = []

            # Build routes by trying to use incoming edge, when it is too short
            if fromEdge._length < 30:
                routes = [k._id + " " + fromEdge._id + " " + toEdge._id for k, v in fromEdge._incoming.items() if
                          all(d._direction not in (d.LINKDIR_TURN, d.LINKDIR_LEFT, d.LINKDIR_RIGHT) for d in v)]

            if not routes:
                routes = [fromEdge._id + " " + toEdge._id]

            extra_routes = []
            # Produce car traffic on the other connections
            for c in node.getConnections():
                if c._direction == c.LINKDIR_TURN:
                    continue

                if c.getFrom() == fromEdge or c.getTo() == toEdge:
                    continue

                r = c.getFrom()._id + " " + c.getTo()._id
                if r not in extra_routes:
                    extra_routes.append(r)

            lanes = [fromEdge._id + "_" + str(i) for i in range(len(fromEdge._lanes))]

            writeRouteFile(p_routes, routes, extra_routes, args.scenario)

            writeDetectorFile(p_detector, "detector", lanes)

            write_scenario(p_scenario, basename(p_network), basename(p_routes), basename(p_detector), args.step_length,
                           time=1800)

            go(p_scenario, args)

            # Read output
            res.append(read_result(folder,
                                   junctionId=node._id,
                                   fromEdgeId=fromEdge._id,
                                   toEdgeId=toEdge._id))

        df = pd.DataFrame(res)
        df.to_csv(join(args.output, "%s.csv" % node._id), index=False)

        print("####################################################################")
        print("[" + str(i) + " / " + str(args.to_index - args.from_index) + "]")


def go(scenario, args):
    traci.start([sumoBinary, "-c", scenario])

    end = int(1800 * (1 / args.step_length))

    try:
        for step in range(0, end):
            traci.simulationStep()
    except Exception as e:
        print(e)

    traci.close()
    sys.stdout.flush()


def setup(parser: ArgumentParser):
    setup_parser(parser)


def main(args):
    args = create_args(args)

    # read in intersections
    with open(args.input[0]) as f:
        selection = set(f.read().splitlines())

    net = sumolib.net.readNet(args.network, withConnections=True, withInternal=False, withFoes=False)
    allNodes = net.getNodes()  # all type of nodes

    traffic_intersections = []
    for node in allNodes:
        if node._type != "internal" and node._id in selection:
            traffic_intersections.append(node)

    print("Total number of junctions:", len(traffic_intersections))

    init_workload(args, traffic_intersections)

    print("Processing: ", args.from_index, ' to ', args.to_index)

    run(args, traffic_intersections)


if __name__ == "__main__":
    parser = ArgumentParser(prog=METADATA[0], description=METADATA[1])
    setup(parser)
    main(parser.parse_args())




import numpy as np
from typing import Dict, Union, Collection, TypeVar

Id = TypeVar('Id', str, int)


class XmlWriter:
    NO_SCOPE = -1
    attributes_current_scope = -2
    ATTRIBUTES_SCOPE_FACTOR = 13
    scope = 0

    def __init__(self, writer):
        self.writer = writer
        self.indent = 0
        self.set_scope(self.NO_SCOPE)

    def _write_line(self, content: str):
        self._write_indent()
        self._write(content + "\n")

    def _write_indent(self):
        self._write("  " * self.indent)

    def _write(self, content: str):
        self.writer.write(bytes(content, "utf-8"))

    def _require_scope(self, scope: int):
        if scope == self.NO_SCOPE and self.scope != self.NO_SCOPE:
            raise RuntimeError("Expected initial scope")

        if not self.scope == scope:
            raise RuntimeError("Expected different scope")

    def set_scope(self, scope: int):
        # print(f"scope {self.scope} -> {scope}")
        self.scope = scope

    def start_attributes(self, expected_scope: int = None):
        if expected_scope is None:  # not checking scope
            expected_scope = self.scope
        self._require_scope(expected_scope)
        self._write_line('<attributes>')
        self.indent += 1
        self.attributes_current_scope = expected_scope * self.ATTRIBUTES_SCOPE_FACTOR
        self.set_scope(self.attributes_current_scope)

    def end_attributes(self):
        self._require_scope(self.attributes_current_scope)
        self.indent -= 1
        self.set_scope(int(self.attributes_current_scope / self.ATTRIBUTES_SCOPE_FACTOR))
        self._write_line('</attributes>')

    def add_attribute(self, name: str, value: Union[str, int, float, bool], typ: str = None):
        if not typ:
            typ = self.get_java_type(type(value))
        self._require_scope(self.attributes_current_scope)
        self._write_line(f'<attribute name="{name}" class="{typ}">{value}</attribute>')

    @staticmethod
    def yes_no(value: bool):
        return "yes" if value else "no"

    @staticmethod
    def true_false(value: bool):
        return "true" if value else "false"

    @staticmethod
    def time(time: int):
        if np.isnan(time):
            return None

        time = int(time)
        hours = time // 3600
        minutes = (time % 3600) // 60
        seconds = (time % 60)
        return f"{hours:02d}:{minutes:02d}:{seconds:02d}"

    @staticmethod
    def get_java_type(python_type: type):
        if 'float' in str(python_type):
            return "java.lang.Double"
        elif 'int64' in str(python_type):
            return "java.lang.Long"
        elif 'int' in str(python_type):
            return "java.lang.Integer"
        elif 'str' in str(python_type):
            return "java.lang.String"
        elif 'bool' in str(python_type):
            return "java.lang.Boolean"
        else:
            return "java.lang.Object"

    def write_preamble_attributes(self, attributes: dict):
        self.start_attributes(expected_scope=0)
        for name, value in attributes.items():
            self.add_attribute(name, value)
        self.end_attributes()


class PopulationWriter(XmlWriter):
    POPULATION_SCOPE = 0
    FINISHED_SCOPE = 1
    PERSON_SCOPE = 2
    PLAN_SCOPE = 3

    def __init__(self, writer):
        XmlWriter.__init__(self, writer)

    def start_population(self, attributes: Dict[str, str] = None):
        self._require_scope(self.NO_SCOPE)
        self._write_line('<?xml version="1.0" encoding="utf-8"?>')
        self._write_line('<!DOCTYPE population SYSTEM "http://www.matsim.org/files/dtd/population_v6.dtd">')
        self._write_line('<population>')
        self.set_scope(self.POPULATION_SCOPE)
        self.indent += 1
        if attributes is not None:
            self.write_preamble_attributes(attributes)

    def end_population(self):
        self._require_scope(self.POPULATION_SCOPE)
        self.indent -= 1
        self._write_line('</population>')
        self.set_scope(self.FINISHED_SCOPE)

    def start_person(self, person_id: Id, attributes: Dict[str, str] = None):
        self._require_scope(self.POPULATION_SCOPE)
        self._write_line(f'<person id="{person_id}">')
        self.set_scope(self.PERSON_SCOPE)
        self.indent += 1
        if attributes:
            self.start_attributes(expected_scope=self.PERSON_SCOPE)
            for k, v in attributes.items():
                self.add_attribute(k, v)
            self.end_attributes()

    def end_person(self):
        self._require_scope(self.PERSON_SCOPE)
        self.indent -= 1
        self.set_scope(self.POPULATION_SCOPE)
        self._write_line('</person>')

    def start_plan(self, selected: bool):
        self._require_scope(self.PERSON_SCOPE)
        self._write_line(f'<plan selected="{self.yes_no(selected)}">')
        self.indent += 1
        self.set_scope(self.PLAN_SCOPE)

    def end_plan(self):
        self._require_scope(self.PLAN_SCOPE)
        self.indent -= 1
        self.set_scope(self.PERSON_SCOPE)
        self._write_line('</plan>')

    def add_activity(self, type: str, x: float, y: float, facility_id: str = None,
                     start_time: int = None, end_time: int = None, max_dur: int = None, attributes: dict = None):
        self._require_scope(self.PLAN_SCOPE)
        self._write_indent()
        self._write('<activity ')
        self._write(f'type="{type}" ')
        self._write(f'x="{x}" y="{y}" ')
        if facility_id is not None: self._write(f'facility="{facility_id}" ')
        if start_time is not None: self._write(f'start_time="{self.time(start_time)}" ')
        if end_time is not None: self._write(f'end_time="{self.time(end_time)}" ')
        if max_dur is not None: self._write(f'max_dur="{self.time(max_dur)}" ')
        if attributes:
            self._write('>\n')
            self.indent += 1
            self.start_attributes(expected_scope=self.PLAN_SCOPE)
            for k, v in attributes.items():
                self.add_attribute(k, v)
            self.end_attributes()
            self.indent -= 1
            self._write_line('</activity>')
        else:
            self._write('/>\n')

    def add_leg(self, mode: str, departure_time: int = None, travel_time: int = None, attributes: dict = None):
        self._require_scope(self.PLAN_SCOPE)
        self._write_indent()
        self._write('<leg ')
        self._write(f'mode="{mode}" ')
        if departure_time is not None: self._write(f'dep_time="{self.time(departure_time)}" ')
        if travel_time is not None: self._write(f'trav_time="{self.time(travel_time)}" ')
        if attributes:
            self._write('>\n')
            self.indent += 1
            self.start_attributes(expected_scope=self.PLAN_SCOPE)
            for k, v in attributes.items():
                self.add_attribute(k, v)
            self.end_attributes()
            self.indent -= 1
            self._write_line('</leg>')
        else:
            self._write('/>\n')


class HouseholdsWriter(XmlWriter):
    HOUSEHOLDS_SCOPE = 0
    FINISHED_SCOPE = 1
    HOUSEHOLD_SCOPE = 2

    def __init__(self, writer):
        XmlWriter.__init__(self, writer)

    def start_households(self, attributes=None):
        self._require_scope(self.NO_SCOPE)
        self._write_line('<?xml version="1.0" encoding="utf-8"?>')
        self._write_line('<households xmlns="http://www.matsim.org/files/dtd" '
                         'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
                         'xsi:schemaLocation="http://www.matsim.org/files/dtd '
                         'http://www.matsim.org/files/dtd/households_v1.0.xsd">')
        self.set_scope(self.HOUSEHOLDS_SCOPE)
        self.indent += 1
        if attributes is not None:
            self.write_preamble_attributes(attributes)

    def end_households(self):
        self._require_scope(self.HOUSEHOLDS_SCOPE)
        self._write_line('</households>')
        self.set_scope(self.FINISHED_SCOPE)

    def start_household(self, household_id: Id):
        self._require_scope(self.HOUSEHOLDS_SCOPE)
        self._write_line(f'<household id="{household_id}">')
        self.set_scope(self.HOUSEHOLD_SCOPE)
        self.indent += 1

    def end_household(self):
        self._require_scope(self.HOUSEHOLD_SCOPE)
        self.indent -= 1
        self.set_scope(self.HOUSEHOLDS_SCOPE)
        self._write_line('</household>')

    def add_members(self, person_ids: Collection[Id]):
        self._require_scope(self.HOUSEHOLD_SCOPE)
        self._write_line('<members>')
        self.indent += 1
        for person_id in person_ids:
            self._write_line(f'<personId refId="{person_id}" />')
        self.indent -= 1
        self._write_line('</members>')

    def add_income(self, income: Union[float, int]):
        self._require_scope(self.HOUSEHOLD_SCOPE)
        self._write_line(f'<income currency="CHF" period="month">{income}</income>')


class FacilitiesWriter(XmlWriter):
    FACILITIES_SCOPE = 0
    FINISHED_SCOPE = 1
    FACILITY_SCOPE = 2

    def __init__(self, writer):
        XmlWriter.__init__(self, writer)

    def start_facilities(self, attributes=None):
        self._require_scope(self.NO_SCOPE)
        self._write_line('<?xml version="1.0" encoding="utf-8"?>')
        self._write_line('<!DOCTYPE facilities SYSTEM "http://www.matsim.org/files/dtd/facilities_v1.dtd">')
        self._write_line('<facilities>')
        self.set_scope(self.FACILITIES_SCOPE)
        self.indent += 1
        if attributes is not None:
            self.write_preamble_attributes(attributes)

    def end_facilities(self):
        self._require_scope(self.FACILITIES_SCOPE)
        self.indent -= 1
        self._write_line('</facilities>')
        self.set_scope(self.FINISHED_SCOPE)

    def start_facility(self, facility_id: Id, x: float, y: float):
        self._require_scope(self.FACILITIES_SCOPE)
        self._write_line(f'<facility id="{facility_id}" x="{x}" y="{y}">')
        self.indent += 1
        self.set_scope(self.FACILITY_SCOPE)

    def end_facility(self):
        self._require_scope(self.FACILITY_SCOPE)
        self.indent -= 1
        self.set_scope(self.FACILITIES_SCOPE)
        self._write_line('</facility>')

    def add_activity(self, purpose: str):
        self._require_scope(self.FACILITY_SCOPE)
        self._write_line(f'<activity type="{purpose}" />')




import pytest
import pathlib

import numpy as np

from matsim import Facility

HERE = pathlib.Path(__file__).parent

files = ['output_facilities.xml.gz']


@pytest.mark.parametrize('filepath', files)
def test_vehicle_reader(filepath):
    facilityReader = Facility.facility_reader(HERE / filepath)
    facilities_dataframe = facilityReader.facilities
    
    facilities_count_types_dataframe = facilities_dataframe['type'].value_counts()
    
    EXPECTED_TOTAL_FACILITIES = 1112
    EXPECTED_EDUCATION_FACILITIES_COUNT = 115
    EXPECTED_HOME_FACILITIES_COUNT = 45
    EXPECTED_OUTSIDE_FACILITIES_COUNT = 404
    EXPECTED_OTHER_FACILITIES_COUNT = 483
    EXPECTED_LEISURE_FACILITIES_COUNT = 13
    EXPECTED_SHOP_FACILITIES_COUNT = 24
    EXPECTED_WORK_FACILITIES_COUNT = 28
    
    EXPECTED_DATAFRAME_KEYS = ['id', 'linkId', 'x', 'y', 'type']
    EXPECTED_ROW_84_RESULT = {
        'id': 'edu_820',
        'linkId': '58693',
        'x': 353742.7,
        'y': 6691356.6,
        'type': 'education'
    }

    assert len(facilities_dataframe) == EXPECTED_TOTAL_FACILITIES
    assert facilities_count_types_dataframe.education == EXPECTED_EDUCATION_FACILITIES_COUNT
    assert facilities_count_types_dataframe.home == EXPECTED_HOME_FACILITIES_COUNT
    assert facilities_count_types_dataframe.outside == EXPECTED_OUTSIDE_FACILITIES_COUNT
    assert facilities_count_types_dataframe.other == EXPECTED_OTHER_FACILITIES_COUNT
    assert facilities_count_types_dataframe.leisure == EXPECTED_LEISURE_FACILITIES_COUNT
    assert facilities_count_types_dataframe.shop == EXPECTED_SHOP_FACILITIES_COUNT
    assert facilities_count_types_dataframe.work == EXPECTED_WORK_FACILITIES_COUNT
    
    assert facilities_dataframe.iloc[84].to_dict() == EXPECTED_ROW_84_RESULT
    np.testing.assert_array_equal(EXPECTED_DATAFRAME_KEYS, facilities_dataframe.keys())


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import re

import sklearn
import sklearn.compose
import sklearn.ensemble
import sklearn.linear_model
import sklearn.svm


try:
    import lightning.regression
    import lightning.classification
except ValueError as e:
    print("Error during lightning import", e)


import xgboost as xgb
import lightgbm as lgb

from sympy.utilities.codegen import codegen


def powerset(iterable):
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))



def create_regressor(trial, classifier_name, random_state):
    """ Create regressor model from given classifier """

    if classifier_name == 'mean':
        model = sklearn.dummy.DummyRegressor()

    elif classifier_name == 'SVR':
        kernel = trial.suggest_categorical('kernel', ['linear', 'poly', 'rbf', 'sigmoid'])
        svc_c = trial.suggest_float('svc_r', 1e-10, 1e10, log=True)
        model = sklearn.svm.SVR(C=svc_c, gamma='auto', kernel=kernel)

    elif classifier_name == 'QLatticeRegressor':
        from .models_feyn import QLatticeRegressor

        model = QLatticeRegressor(trial.suggest_int('complexity', 3, 6, step=1), random_state=random_state)

    elif classifier_name == 'RandomForestRegressor':
        rf_max_depth = trial.suggest_int('rf_max_depth', 2, 4)
        rf_n = trial.suggest_int('rf_n', 5, 30, step=5)
        min_samples_split = trial.suggest_int('min_samples_split', 2, 8, step=1)
        b = trial.suggest_categorical('bootstrap', [True, False])

        model = sklearn.ensemble.RandomForestRegressor(max_depth=rf_max_depth, n_estimators=rf_n,
                                                       oob_score=b, min_samples_split=min_samples_split,
                                                       random_state=random_state)
    elif classifier_name == 'XGBRFRegressor':

        params = {
            'objective': 'reg:squarederror',
            'eval_metric': 'mae',
            'max_depth': trial.suggest_int('max_depth', low=2, high=4),
            'min_child_weight': trial.suggest_int('min_child_weight', low=1, high=10),
            'eta': trial.suggest_float('eta', 0.01, 1, log=True),
            'lambda': trial.suggest_float('lambda', 0.01, 1, log=True),
            'alpha': trial.suggest_float('alpha', 0.01, 1, log=True),
            'gamma': trial.suggest_float('gamma', 0.01, 1, log=True),
            'n_estimators': trial.suggest_int('rf_n', 5, 30, step=5),
            'subsample': 0.9,
            'colsample_bytree': 0.9,
            'colsample_bynode': 0.9,
            'random_state': random_state
        }

        model = xgb.XGBRFRegressor(**params)

    elif classifier_name == 'XGBRegressor':

        params = {
            'objective': 'reg:squarederror',
            'eval_metric': 'mae',
            'max_depth': trial.suggest_int('max_depth', low=2, high=4),
            'min_child_weight': trial.suggest_int('min_child_weight', low=1, high=10),
            'eta': trial.suggest_float('eta', 0.01, 1, log=True),
            'lambda': trial.suggest_float('lambda', 0.01, 1, log=True),
            'alpha': trial.suggest_float('alpha', 0.01, 1, log=True),
            'gamma': trial.suggest_float('gamma', 0.01, 1, log=True),
            'n_estimators': trial.suggest_int('rf_n', 5, 30, step=5),
            'subsample': 0.9,
            'colsample_bytree': 0.9,
            'colsample_bynode': 0.9,
            'random_state': random_state
        }

        model = xgb.XGBRegressor(**params)

    elif classifier_name == 'ExtraTreesRegressor':

        rf_n = trial.suggest_int('rf_n', 5, 30, step=5)
        min_samples_split = trial.suggest_int('min_samples_split', 2, 8, step=1)
        max_depth = trial.suggest_int('max_depth', 2, 4)
        b = trial.suggest_categorical('bootstrap', [True, False])

        model = sklearn.ensemble.ExtraTreesRegressor(n_estimators=rf_n,
                                                     min_samples_split=min_samples_split,
                                                     max_depth=max_depth,
                                                     bootstrap=b, oob_score=b,
                                                     random_state=random_state)

    elif classifier_name == 'LGBMRegressor':

        params = {
            'objective': 'regression',
            'num_leaves': trial.suggest_int('num_leaves', 2, 32, log=True),
            'n_estimators': trial.suggest_int('n_estimators', 5, 30, step=5),
            #  'min_child_weight': trial.suggest_loguniform('min_child_weight', 0.0001, 1),
            #  'min_child_samples': trial.suggest_int('min_child_samples', low=5, high=30, step=5),
            'subsample': 0.9,
            'subsample_freq': 10,
            'colsample_bytree': 0.9,
            'reg_alpha': trial.suggest_float('alpha', 1e-10, 1, log=True),
            'random_state': random_state
        }

        model = lgb.LGBMRegressor(**params)

    elif classifier_name == "DecisionTreeRegressor":

        min_samples_split = trial.suggest_int('min_samples_split', 2, 8, step=1)
        max_depth = trial.suggest_int('max_depth', 2, 4)

        model = sklearn.tree.DecisionTreeRegressor(max_depth=max_depth, min_samples_split=min_samples_split,
                                                   random_state=random_state)


    elif classifier_name == "BaggingRegressor":

        reg = sklearn.linear_model.PassiveAggressiveRegressor(C=trial.suggest_float('C', 0.0001, 100, log=True))
        model = sklearn.ensemble.BaggingRegressor(reg,
                                                  n_estimators=trial.suggest_int('rf_n', 5, 30, step=5),
                                                  max_samples=trial.suggest_float('alpha', 0.2, 1),
                                                  random_state=random_state)

    elif classifier_name == 'Ridge':

        model = sklearn.linear_model.Ridge(alpha=trial.suggest_float('alpha', 0.001, 1000, log=True), solver='saga',
                                           random_state=random_state)

    elif classifier_name == 'Lasso':

        model = sklearn.linear_model.Lasso(alpha=trial.suggest_float('alpha', 0.001, 1000, log=True),
                                           random_state=random_state)


    elif classifier_name == 'ElasticNet':

        model = sklearn.linear_model.ElasticNet(alpha=trial.suggest_float('alpha', 0.01, 1, log=True),
                                                l1_ratio=trial.suggest_float('ratio', 0, 1),
                                                random_state=random_state)
    elif classifier_name == 'SGDRegressor':

        model = sklearn.linear_model.SGDRegressor(alpha=trial.suggest_float('alpha', 0.0001, 1, log=True),
                                                  random_state=random_state)

    elif classifier_name == "LinearSVR":

        model = sklearn.svm.LinearSVR(C=trial.suggest_float('C', 0.001, 1000, log=True), dual=False, max_iter=2000,
                                      loss="squared_epsilon_insensitive", random_state=random_state)

    elif classifier_name == 'PassiveAggressiveRegressor':

        model = sklearn.linear_model.PassiveAggressiveRegressor(C=trial.suggest_float('C', 0.0001, 100, log=True),
                                                                random_state=random_state)

    elif classifier_name == 'LogisticRegression':

        model = sklearn.linear_model.LogisticRegression(C=trial.suggest_float('C', 0.0001, 100, log=True),
                                                        solver='saga', random_state=random_state)

    elif classifier_name == 'AdaGradRegressor':

        model = lightning.regression.AdaGradRegressor(eta=trial.suggest_float('eta', 0.0001, 10, log=True),
                                                      alpha=trial.suggest_float('alpha', 0.0001, 10, log=True),
                                                      random_state=random_state)

    elif classifier_name == 'CDClassifier':

        model = lightning.regression.CDClassifier(C=trial.suggest_float('C', 0.0001, 10, log=True),
                                                  alpha=trial.suggest_float('alpha', 0.0001, 10, log=True),
                                                  random_state=random_state)

    elif classifier_name == 'FistaClassifier':

        model = lightning.regression.FistaClassifier(C=trial.suggest_float('C', 0.0001, 10, log=True),
                                                     alpha=trial.suggest_float('alpha', 0.0001, 10, log=True),
                                                     random_state=random_state)

    elif classifier_name == 'SDCAClassifier':

        model = lightning.regression.SDCAClassifier(alpha=trial.suggest_float('alpha', 0.0001, 10, log=True),
                                                    random_state=random_state)

    elif classifier_name == 'OneClassSVM':

        kernel = trial.suggest_categorical('kernel', ['linear', 'poly', 'rbf', 'sigmoid'])
        gamma = trial.suggest_categorical('gamma', ['scale', 'auto'])
        nu = trial.suggest_float('nu', 0, 1)

        model = sklearn.svm.OneClassSVM(kernel=kernel, gamma=gamma, nu=nu, random_state=random_state)

    elif classifier_name == 'KernelSVC':

        kernel = trial.suggest_categorical('kernel', ['linear', 'poly', 'rbf', 'sigmoid', 'cosine'])
        model = lightning.regression.KernelSVC(alpha=trial.suggest_float('alpha', 0.0001, 10, log=True),
                                               kernel=kernel, random_state=random_state)

    else:
        raise Exception("Unknown regressor: " + classifier_name)

    return model


def sympy_to_c(model):
    # classifier can also be used directly if only 2 classes
    [(c_name, c_code), (h_name, c_header)] = codegen(("score", model.sympify()), "C89", "model", header=False,
                                                     empty=False)
    c_code = re.sub(r"data([0-9]+)", r"data[\1]", c_code)

    return c_code


def model_to_java(name, package, model, scaler, df):
    """ Convert to java source file """
    import m2cgen as m2c

    code = m2c.export_to_java(model, package, name)

    code, params = replace_params(code)

    imp = """import org.matsim.application.prepare.network.opt.FeatureRegressor;
import it.unimi.dsi.fastutil.objects.Object2DoubleMap;
    
/**
* Generated model, do not modify.
*/
public final class"""

    code = code.replace("public class", imp)
    code = code.replace(name, name + " implements FeatureRegressor")

    features = [f"data[{i}] = {s};\n" for i, s in enumerate(model_features(scaler, df))]

    idx = code.index("public static double score")

    pre = """
    public static %s INSTANCE = new %s();
    public static final double[] DEFAULT_PARAMS = %s;

    @Override
    public double predict(Object2DoubleMap<String> ft) {
        return predict(ft, DEFAULT_PARAMS);
    }
    
    @Override
    public double[] getData(Object2DoubleMap<String> ft) {
        double[] data = new double[%d];
""" % (name, name, str(params).replace("[", "{").replace("]", "}"), len(features))

    for ft in features:
        pre += "\t\t" + ft

    pre += """
        return data;
    }
    
    @Override
    public double predict(Object2DoubleMap<String> ft, double[] params) {

        double[] data = getData(ft);
        for (int i = 0; i < data.length; i++)
            if (Double.isNaN(data[i])) throw new IllegalArgumentException("Invalid data at index: " + i);
    
        return score(data, params);
    }
    """

    code = code.replace("score(double[] input)", "score(double[] input, double[] params)")

    return f"{code[:idx]}{pre}{code[idx:]}"


def model_to_py(name, model, scaler, df):
    import m2cgen as m2c

    code = m2c.export_to_python(model, function_name="score")

    code, params = replace_params(code)

    code = code.replace("input", "inputs")
    code = code.replace("(inputs):", "(params, inputs):")

    features = "\t\t".join([f"data[{i}] = {s}\n" for i, s in enumerate(model_features(scaler, df))])
    features = features.replace(".getDouble(", ".get(")

    code = """def features(ft, data):
\t\t%s
params = %s
""" % (features, params) + code

    code += """
def batch_loss(params, inputs, targets):
    error = 0
    for x, y in zip(inputs, targets):
        preds = score(params, x)
        error += (preds - y) ** 2
    return error
"""

    return code


def replace_params(code):
    """ Replaces and collects model parameters """

    pattern = re.compile(r"(var\d+) = (-?\d+.\d+)")

    params = []
    new_code = ''
    start = 0

    for m in pattern.finditer(code):
        end, newstart = m.span()
        new_code += code[start:end]
        rep = "%s = params[%d]" % (m.group(1), len(params))
        new_code += rep
        start = newstart

        params.append(float(m.group(2)))

    new_code += code[start:]

    return new_code, params


def model_features(scaler, df):
    for name, t, ids in scaler.transformers_:

        for i, idx in enumerate(ids):

            c = df.columns[idx]

            if name == "scale":

                t = scaler.named_transformers_[name]

                with_mean = t.get_params()["with_mean"]

                if with_mean:
                    yield f"(ft.getDouble(\"{c}\") - {t.mean_[i]}) / {t.scale_[i]}"
                else:
                    yield f"ft.getDouble(\"{c}\") / {t.scale_[i]}"

            elif t == "passthrough":
                yield f"ft.getDouble(\"{c}\")"

            elif t == "drop":
                continue

            else:
                raise Exception("Unknown transformer: " + t)



import pytest
import pathlib

from collections import defaultdict

from matsim import Events

HERE = pathlib.Path(__file__).parent

actTypes = {'actstart', 'actend', 'departure', 'arrival', 'travelled', 'PersonEntersVehicle', 'PersonLeavesVehicle',
            'vehicle enters traffic', 'vehicle leaves traffic', 'left link', 'entered link'}

files = ['output_events.xml.gz', 'output_events.pb.gz', 'output_events.ndjson.gz']


@pytest.mark.parametrize('filepath', files)
def test_event_reader(filepath):
    events = Events.event_reader(HERE / filepath)
    count = defaultdict(int)

    for event in events:
        count[event['type']] += 1

    assert len(count) == 10
    assert count['actend'] == 201
    assert count['departure'] == 201
    assert count['PersonEntersVehicle'] == 201
    assert count['vehicle enters traffic'] == 201
    assert count['left link'] == 700
    assert count['entered link'] == 700
    assert count['vehicle leaves traffic'] == 201
    assert count['PersonLeavesVehicle'] == 201
    assert count['arrival'] == 201
    assert count['actstart'] == 201


@pytest.mark.parametrize('filepath', files)
def test_event_filter_commas(filepath):
    events = Events.event_reader(HERE / filepath, types='actend,left link')
    count = defaultdict(int)

    for event in events:
        count[event['type']] += 1

    assert len(count) == 2
    assert count['actend'] == 201
    assert count['left link'] == 700


@pytest.mark.parametrize('filepath', files)
def test_event_filter(filepath):
    events = Events.event_reader(HERE / filepath, types={'actend', 'departure', 'left link'})
    count = defaultdict(int)

    for event in events:
        count[event['type']] += 1

    assert len(count) == 3
    assert count['actend'] == 201
    assert count['departure'] == 201
    assert count['left link'] == 700


def test_non_existent():
    with pytest.raises(IOError):
        events = Events.event_reader("not existing.xml")
        for _ in events:
            pass



# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from optuna.trial import TrialState


def study_as_df(study):
    """ Convert study to dataframe """
    completed = completed_trials(study)

    # modes = study.user_attrs["modes"]
    # fixed_mode = study.user_attrs["fixed_mode"]

    data = []

    for i, trial in enumerate(completed):

        entry = {
            "trial": i,
            "start": trial.datetime_start,
            "duration": trial.duration,
        }

        for k, v in trial.user_attrs.items():
            if type(v) in (float, int):
                entry[k] = v

        data.append(entry)

    return pd.DataFrame(data)


def default_chain_scheduler(completed):
    """ Default function to determine when to chain runs. """

    n = len(completed)

    if n <= 6:
        return n % 2 == 0

    if n <= 15:
        return n % 5 == 0

    if n <= 50:
        return n % 10 == 0

    return False


def linear_scheduler(start=0.6, end=1, interval=3):
    """ Creates a lr scheduler that will interpolate linearly from start to end over the first n iterations.

        :param start: Initial learning rate.
        :param end: Final learning rate to reach.
        :param interval: Number of runs until end rate should be reached.
    """
    if interval < 2:
        raise ValueError("N must be greater or equal 2.")

    def _fn(n, *args, **kwargs):

        if n > interval:
            return end

        return start + (n - 1) * (end - start) / interval

    return _fn


def completed_trials(study):
    completed = filter(lambda s: s.state == TrialState.COMPLETE, study.trials)
    return sorted(completed, key=lambda s: s.number)


def same_sign(x):
    x = x.to_numpy()
    return np.all(x >= 0) if x[0] >= 0 else np.all(x < 0)


def cli_oldstyle(yaml_arg="params"):
    """ Produces CLI for non matsim application scenarios  """

    def _f(jvm_args, jar, config, params_path, run_dir, trial_number, run_args):
        return "java %s -jar %s %s --config:controler.outputDirectory %s --config:controler.runId %03d --%s %s %s" % (
            jvm_args, jar, config, run_dir, trial_number, yaml_arg, params_path, run_args
        )

    return _f



#!/usr/bin/env python
# @author  Angelo Banse, Ronald Nippold, Christian Rakow

import os
import sys
from argparse import ArgumentParser
from os.path import join, basename
from traceback import print_exc

from .utils import init_workload, setup_parser, create_args, write_scenario, filter_network_polygon, vehicle_parameter
from .randomTrips import get_options, main as gen_trips

import sumolib.net
import traci  # noqa

from sumolib import checkBinary  # noqa

import lxml.etree as ET

import pandas as pd

sumoBinary = checkBinary('sumo')
netconvert = checkBinary('netconvert')

METADATA = "sumo-routes", "Determine avg. uncongested link speed with SUMO."


def writeRouteFile(f_name, fromEdge, toEdge, veh, end, scenario):
    """ Write route file using flow """
    # https://sumo.dlr.de/docs/Demand/Shortest_or_Optimal_Path_Routing.html
    # https://sumo.dlr.de/docs/Definition_of_Vehicles%2C_Vehicle_Types%2C_and_Routes.html#incomplete_routes_trips_and_flows

    text = f"""<?xml version="1.0" encoding="UTF-8"?>
<routes xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/routes_file.xsd">
    <vTypeDistribution id="vDist">
        {vehicle_parameter(scenario)}
    </vTypeDistribution>
    <flow id="veh" begin="0" end="{end}" vehsPerHour="{veh}" type="vDist" from="{fromEdge}" to="{toEdge}" departSpeed="speedLimit"/>
    
</routes>
"""

    with open(f_name, "w") as f:
        f.write(text)


def writeDetectorFile(f_name, begin):
    """ Write files needed for analysis """

    # https://sumo.dlr.de/docs/Simulation/Output/Lane-_or_Edge-based_Traffic_Measures.html

    text = f"""<?xml version="1.0" encoding="UTF-8"?>
	<additional xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/additional_file.xsd">
	    <edgeData id="out" file="out.xml" begin="%d" excludeEmpty="true"/>
	</additional>
	""" % begin

    with open(f_name, 'w') as f:
        f.write(text)


def read_result(out):
    """ Reads result from output xml """

    data = []

    for _, elem in ET.iterparse(out, events=("end",),
                                tag=('edge',),
                                remove_blank_text=True):
        d = {
            "edgeId": elem.attrib["id"]
        }
        for a in ("traveltime", "density", "waitingTime", "timeLoss", "speed", "speedRelative"):
            d[a] = float(elem.attrib.get(a, float("nan")))

        # Skip the non primary routes
        if int(elem.attrib.get("entered", "0")) < 100 and int(elem.attrib.get("left", "0")) < 100:
            continue

        data.append(d)

    return pd.DataFrame(data)


def run(args, routes, location_offset):
    i = 0

    if args.to_index <= 0:
        args.to_index = len(routes)

    for x in range(args.from_index, args.to_index):
        route = routes.iloc[x]
        i += 1

        print(route)

        p_network = join(args.runner, "filtered.net.xml")
        p_routes = join(args.runner, "route.rou.xml")
        p_detector = join(args.runner, "detector.add.xml")
        p_trips = join(args.runner, "trips.trips.xml")

        # 1hour simulation plus travel time
        end = int(route.travel_time + 3600)

        filter_network_polygon(netconvert, args.network, location_offset, route.geometry, p_network)

        # Nearly uncongested vehicle flow
        writeRouteFile(p_routes, route.fromEdge, route.toEdge, int(route.min_capacity * 0.3), end, args.scenario)
        writeDetectorFile(p_detector, route.travel_time)

        # Produce some very light traffic on the other roads
        gen_trips(get_options(["-n", p_network, "-o", p_trips, "-r", join(args.runner, "random_routes.rou.xml"),
                          "--validate", "-e", end, '-t type="vDist"',
                          "--insertion-density", args.insertion_density, "--fringe-factor", "max"]))

        p_scenario = join(args.runner, "scenario.sumocfg")

        write_scenario(p_scenario, basename(p_network), basename(p_routes) + "," + "trips.trips.xml",
                       basename(p_detector), args.step_length, end)

        try:
            go(p_scenario, p_network, end, route.fromEdge + "_" + route.toEdge, args)
        except Exception as e:
            print_exc()

        print("####################################################################")
        print("[" + str(i) + " / " + str(args.to_index - args.from_index) + "]")


def go(scenario, network, end, f, args):
    # Clean existing output
    out = join(args.runner, "out.xml")

    if os.path.exists(out):
        os.remove(out)

    traci.start([sumoBinary, "-n", network], port=args.port)

    steps = int(end * (1 / args.step_length))

    # Load scenario with desired traffic scaling
    traci.load(["-c", scenario])
    try:
        for step in range(0, steps):
            traci.simulationStep()
    except Exception as e:
        print(e)

    traci.close()

    res = read_result(out)
    res.to_csv(join(args.output, f + ".csv"), index=False)

    sys.stdout.flush()


def setup(parser: ArgumentParser):
    setup_parser(parser)

    parser.add_argument("--insertion-density", type=float, default=10, help="Density of background vehicles")


def main(args):

    args = create_args(args)

    # Net is needed to get the bounds
    net = sumolib.net.readNet(args.network, withConnections=False, withInternal=False, withFoes=False)

    df = pd.read_csv(args.input[0])

    init_workload(args, df)

    print("Total number of routes:", len(df))
    print("Processing: ", args.from_index, ' to ', args.to_index)

    run(args, df, net.getLocationOffset())


if __name__ == "__main__":
    parser = ArgumentParser(prog=METADATA[0], description=METADATA[1])
    setup(parser)
    main(parser.parse_args())




import numpy as np

from matsim.calibration import ASCCalibrator


def test_asc_sampler():
    # variables are not needed for this test
    sampler = ASCCalibrator([], {}, {})

    np.random.seed(0)

    utils = np.random.rand(1000, 4)
    ascs = np.zeros(4)

    # real shares (target)
    z = [0.1, 0.5, 0.3, 0.1]

    for it in range(200):

        modes = np.argmax(utils + ascs + np.random.normal(size=(1000, 4)), axis=1)
        _, counts = np.unique(modes, return_counts=True)

        share = counts / np.sum(counts)

        for i in range(1, len(share)):
            ascs[i] += sampler.calc_asc_update(z[i], share[i], z[0], share[0])

        err = np.sum(np.abs(z - share))

        print(share, ascs, err)

    assert err < 0.12


if __name__ == "__main__":
    test_asc_sampler()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: matsim/pb/Ids.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13matsim/pb/Ids.proto\x12\tmatsim.pb\"E\n\x07ProtoId\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x05\x12\x1f\n\x04type\x18\x03 \x01(\x0e\x32\x11.matsim.pb.IdType*\x9b\x01\n\x06IdType\x12\x0e\n\nID_UNKNOWN\x10\x00\x12\x10\n\x0cID_DEPARTURE\x10\x01\x12\x14\n\x10ID_TRANSIT_ROUTE\x10\x02\x12\x13\n\x0fID_TRANSIT_LINE\x10\x03\x12\x0e\n\nID_VEHICLE\x10\x04\x12\x0b\n\x07ID_LINK\x10\x05\x12\r\n\tID_PERSON\x10\x06\x12\x18\n\x14ID_ACTIVITY_FACILITY\x10\x07\x42\x1c\n\x18org.matsim.core.utils.pbP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'matsim.pb.Ids_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030org.matsim.core.utils.pbP\001'
  _globals['_IDTYPE']._serialized_start=106
  _globals['_IDTYPE']._serialized_end=261
  _globals['_PROTOID']._serialized_start=34
  _globals['_PROTOID']._serialized_end=103
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python

import os
import sys
from subprocess import call

from shapely import wkt
from shapely.ops import transform

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)

def setup_parser(parser):
    parser.add_argument("input", nargs=1, help="Path to input csv")

    parser.add_argument("--output", default="output", help="Path to output folder")
    parser.add_argument("--network", type=str, default="../../../../scenarios/input/sumo.net.xml",
                        help="Path to network file")
    parser.add_argument("--veh", type=int, default=5000, help="Vehicles per hour per lane to simulate")
    parser.add_argument("--scenario", type=str, default="base", choices=["base", "sst", "st", "mt", "lt"],
                        help="Name of scenario for vehicle share and capabilities")
    parser.add_argument("--from-index", type=int, default=0, help="Start from number")
    parser.add_argument("--to-index", type=int, default=-1, help="Stop at number")
    parser.add_argument("--step-length", type=float, default=0.2, help="SUMO step length")
    parser.add_argument("--runner", type=str, default="runner0", help="Runner name")
    parser.add_argument("--runner-total", type=int, default=0, help="Total number of runners")
    parser.add_argument("--runner-index", type=int, default=0, help="Runner index")


def create_args(args):
    import sumolib

    args.port = sumolib.miscutils.getFreeSocketPort()

    os.makedirs(args.output, exist_ok=True)
    os.makedirs(args.runner, exist_ok=True)

    return args

def vehicle_parameter(scenario):
    """ Predefined scenarios for vehicle parameters """

    if scenario == "base":
        return """
            <vType id="vehCV" probability="0.99" color="1,0,0" vClass="passenger" impatience="0.1"/>
            <vType id="vehAV" probability="0.01" color="0,1,0" vClass="passenger" decel="3.0" sigma="0.1" tau="1.5" speedFactor="1" speedDev="0" />
        """

    elif scenario == "sst":
        return """
            <vType id="vehCV" probability="0.9" color="1,0,0" vClass="passenger" impatience="0.1"/>
            <vType id="vehAV" probability="0.1" color="0,1,0" vClass="passenger" decel="3.0" sigma="0.1" tau="1.3" speedFactor="1" speedDev="0" />
        """

    elif scenario == "st":  # 5-10 years
        return """
            <vType id="vehCV" probability="0.7" color="1,0,0" vClass="passenger" impatience="0.1"/>
            <vType id="vehAV" probability="0.3" color="0,1,0" vClass="passenger" decel="3.0" sigma="0.1" tau="1.2" speedFactor="1" speedDev="0" />
        """
    elif scenario == "mt":  # 15-20 years
        return """
            <vType id="vehCV" probability="0.3" color="1,0,0" vClass="passenger" impatience="0.1"/>
            <vType id="vehAV" probability="0.55" color="0,1,0" vClass="passenger" decel="4.5" sigma="0.05" tau="0.9" speedFactor="1" speedDev="0" />
            <vType id="vehACV" probability="0.15" color="0,0,1" vClass="passenger" minGap="1" accel="2.6" decel="4.5" sigma="0.05" tau="0.8" speedFactor="1" speedDev="0" impatience="0"/>
        """
    elif scenario == "lt":  # 25+ years
        return """
            <vType id="vehCV" probability="0.05" color="1,0,0" vClass="passenger" impatience="0.1"/>
            <vType id="vehAV" probability="0.15" color="0,1,0" vClass="passenger" decel="4.5" sigma="0.05" tau="0.9" speedFactor="1" speedDev="0" />
            <vType id="vehACV" probability="0.8" color="0,0,1" vClass="passenger" minGap="0.5" accel="2.6" decel="4.5" sigma="0" tau="0.6" speedFactor="1" speedDev="0" impatience="0.8"/>
    """

    raise Exception("Unknown scenario: " + scenario)


def init_workload(args, items):
    """ Set indices for the runner automatically """
    if args.runner_total <= 1:
        return

    n = len(items)
    step = n // args.runner_total

    args.from_index = args.runner_index * step
    args.to_index = min((args.runner_index + 1) * step, n)


def write_scenario(f, network_file, route_file, additional_file, step_length=0.2, time=600):
    """ Write sumo scenario file """

    with open(f, "w") as fn:
        fn.write("""<?xml version="1.0" encoding="UTF-8"?>
<configuration xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://sumo.dlr.de/xsd/sumoConfiguration.xsd">

    <input>
        <net-file value="%s"/>
        <route-files value="%s"/>
        <additional-files value="%s"/>
    </input>

    <time>
        <begin value="0"/>
        <end value="%d"/>
        <step-length value="%f"/>
    </time>
</configuration>
""" % (network_file, route_file, additional_file, time, step_length))


def filter_network(netconvert, netfile, edge, output, args=None):
    if isinstance(edge, list):
        x = [s[0] for e in edge for s in e.getShape()]
        y = [s[1] for e in edge for s in e.getShape()]
    else:
        x = [s[0] for s in edge.getShape()]
        y = [s[1] for s in edge.getShape()]

    # minX,minY,maxX,maxY
    boundary = ",".join(str(s) for s in [min(x) - 50, min(y) - 50, max(x) + 50, max(y) + 50])

    cmd = [netconvert, '-s', netfile, "--keep-edges.in-boundary", boundary]

    if args:
        cmd += args

    cmd += ['-o', output]

    call(cmd)


def filter_network_polygon(netconvert, netfile, location_offset, geometry, output):
    """ Filter network with a list of polygon coordinates"""

    polygon = wkt.loads(geometry)

    polygon = transform(lambda x, y: (x + location_offset[0], y + location_offset[1]), polygon)

    coords = ",".join("%.2f,%.2f" % f for f in polygon.exterior.coords)

    cmd = [netconvert, '-s', netfile, "--keep-edges.in-boundary", coords, "--no-internal-links", "false"]
    cmd += ['-o', output]

    call(cmd)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

from .data import run_create_ref_data
from .data import run_extract_activities
from .data import run_lookup_regiostar
from .network import run_collect_results
from .network import run_edges as sumo_edges
from .network import run_intersections as sumo_intersections
from .network import run_routes as sumo_routes
from .network import run_train_model


def _add(subparsers, m):
    """ Adds module to as subcommand"""
    s1 = subparsers.add_parser(m.METADATA[0], help=m.METADATA[1])
    m.setup(s1)
    s1.set_defaults(func=m.main)


def main():
    """ Main entry point. """

    parser = argparse.ArgumentParser(prog='matsim-scenariogen',
                                     description="MATSim scenario command line utility")

    subparsers = parser.add_subparsers(title="Subcommands")

    _add(subparsers, sumo_edges)
    _add(subparsers, sumo_routes)
    _add(subparsers, sumo_intersections)
    _add(subparsers, run_train_model)
    _add(subparsers, run_collect_results)

    try:
        from .network import run_opt_freespeed
        _add(subparsers, run_opt_freespeed)
    except Exception as e:
        print("Opt freespeed not available", e)

    _add(subparsers, run_extract_activities)
    _add(subparsers, run_create_ref_data)
    _add(subparsers, run_lookup_regiostar)

    args = parser.parse_args()

    if not hasattr(args, 'func'):
        parser.print_help()
    else:
        args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import ArgumentParser

from . import clean_iters as ci

def main():
    """ Main entry point. """

    parser = ArgumentParser(prog='matsim-tools', description="MATSim command line util")
    subparsers = parser.add_subparsers(title="Subcommands")

    s1 = subparsers.add_parser(ci.METADATA[0], help=ci.METADATA[1])
    ci.setup(s1)
    s1.set_defaults(func=ci.main)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

from . import read_all
from .preparation import prepare_persons, create_activities

METADATA = "data-extract-activities", "Extract activities and persons from survey data."


def setup(parser: argparse.ArgumentParser):
    parser.add_argument("dirs", nargs="+", help="Directories with survey data")
    parser.add_argument("--regiostar", default=None)

    parser.add_argument("--output", default="table", help="Output prefix")


def main(args):
    hh, persons, trips = read_all(args.dirs, regio=args.regiostar)

    hh.to_csv(args.output + "-households.csv")
    trips.to_csv(args.output + "-trips.csv")
    persons.to_csv(args.output + "-unscaled-persons.csv")

    print("Written survey csvs")

    df = prepare_persons(hh, persons, trips, augment=5, core_weekday=True, remove_with_invalid_trips=True)

    df.to_csv(args.output + "-persons.csv", index_label="idx")
    print("Created %d synthetics persons" % len(df))

    activities = create_activities(df, trips, include_person_context=False, cut_groups=False)
    activities.to_csv(args.output + "-activities.csv", index=False)



import xopen
import xml.etree.ElementTree as ET
import pandas as pd
from matsim import utils

class Vehicle:
    def __init__(self, vehicle_types, vehicles):
        self.vehicle_types = vehicle_types
        self.vehicles = vehicles

# Returns vehicle_types and vehicles dataframes
# <vehicleType> attributes and children attributes are récursively added to the dataframe
def vehicle_reader(filename, convert_dataframes_types=True):
    tree = ET.iterparse(xopen.xopen(filename, 'r'), events=['start','end'])
    
    vehicle_types = []
    vehicles = []
    
    current_vehicle_type = {}
    current_vehicle = {}
    
    is_parsing_vehicle_type = False
    
    for xml_event, elem in tree:
        _, _, elem_tag = elem.tag.partition('}')     # Removing xmlns tag from tag name
        
        # VEHICLES
        if elem_tag == 'vehicle' and xml_event == 'start':
            utils.parse_attributes(elem, current_vehicle)
        
        elif elem_tag == 'vehicle' and xml_event == 'end':
            vehicles.append(current_vehicle)
            current_vehicle = {}
            elem.clear()
            
        # VEHICLETYPES
        elif elem_tag == 'vehicleType' and xml_event == 'start':
            utils.parse_attributes(elem, current_vehicle_type)
            is_parsing_vehicle_type = True
        
        # ATTRIBUTES
        elif elem_tag == 'attribute' and xml_event == 'start':
            current_vehicle_type[elem.attrib['name']] = elem.text
        
        # LENGTH / WIDTH
        elif elem_tag in ['length', 'width'] and xml_event == 'start':
            current_vehicle_type[elem_tag] = elem.attrib['meter']
        
        # VEHICLETYPES
        elif elem_tag == 'vehicleType' and xml_event == 'end':
            vehicle_types.append(current_vehicle_type)
            current_vehicle_type = {}
            elem.clear()
            is_parsing_vehicle_type = False
        
        # EVERYTHING ELSE
        elif is_parsing_vehicle_type and elem_tag not in ['attribute', 'length', 'width']:
            utils.parse_attributes(elem, current_vehicle_type)

    
    vehicle_types = pd.DataFrame.from_records(vehicle_types)
    vehicles = pd.DataFrame.from_records(vehicles)
    
    if convert_dataframes_types:
        try:
            vehicle_types['accessTimeInSecondsPerPerson'] = vehicle_types['accessTimeInSecondsPerPerson'].astype(float)
            vehicle_types['egressTimeInSecondsPerPerson'] = vehicle_types['egressTimeInSecondsPerPerson'].astype(float)
            vehicle_types['seats'] = vehicle_types['seats'].astype(int)
            vehicle_types['standingRoomInPersons'] = vehicle_types['standingRoomInPersons'].astype(int)
            vehicle_types['length'] = vehicle_types['length'].astype(float)
            vehicle_types['width'] = vehicle_types['width'].astype(float)
            vehicle_types['pce'] = vehicle_types['pce'].astype(float)
            vehicle_types['factor'] = vehicle_types['factor'].astype(float)
        except KeyError:
            print('dataframe types conversion failed')
    
    return Vehicle(vehicle_types, vehicles)
        


import pytest
import pathlib
import numpy as np
from matsim import Household

HERE = pathlib.Path(__file__).parent

files = ['output_households.xml.gz']

@pytest.mark.parametrize('filepath', files)
def test_vehicle_reader(filepath):
    household_reader = Household.houshold_reader(HERE / filepath)
    
    household_dataframe = household_reader.households
    
    EXPECTED_TOTAL_HOUSEHOLDS = 112
    EXPECTED_HOUSEHOLD_COLUMNS = ['id', 'bikeAvailability', 'carAvailability', 'censusId', 'household_income', 'members']
    EXPECTED_ROW_84_RESULT = {
        'id': 2293,
        'bikeAvailability': 'none',
        'carAvailability': 'some',
        'censusId': 503,
        'household_income': 3916.708525944572,
        'members' : [5427, 5428, 5429, 5430, 5431]
    }
    
    
    assert len(household_dataframe) == EXPECTED_TOTAL_HOUSEHOLDS
    assert household_dataframe.iloc[84].to_dict() == EXPECTED_ROW_84_RESULT
    np.testing.assert_array_equal(EXPECTED_HOUSEHOLD_COLUMNS, household_dataframe.keys())


class TripEventHandler:

    def __init__(self, main_mode_identifier, person_filter):
        self.drivers = set()
        self.persons = {}
        self.stuck = set()
        self.main_mode_identifier = main_mode_identifier
        self.person_filter = person_filter
        self.handlers = {
            'actstart': self.on_activity_start, 'actend': self.on_activity_end, 'departure': self.on_person_departure,
            'arrival': self.on_person_arrival, 'stuck': self.on_person_stuck
        }

    def on_event(self, time, type, attrs):
        if (type in self.handlers):
            self.handlers[type](time, attrs)

    def on_transit_driver_starts(self, time, attrs):
        self.drivers.add(attrs['id'])

    def on_activity_end(self, time, attrs):
        if self.is_stage_activity(attrs['acttype']) or attrs['id'] in self.drivers or not self.person_filter(
                attrs['id']):
            return

    def on_activity_start(self, time, attrs):
        pass

    def on_person_arrival(self, time, attrs):
        pass

    def on_person_departure(self, time, attrs):
        pass

    def on_person_stuck(self, time, attrs):
        pass

    def is_stage_activity(self, type):
        return type.endswith("interaction")





# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: matsim/pb/Model.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15matsim/pb/Model.proto\x12\tmatsim.pb\"\"\n\nCoordinate\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x42&\n\x18org.matsim.core.utils.pbB\nProtoModelb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'matsim.pb.Model_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030org.matsim.core.utils.pbB\nProtoModel'
  _globals['_COORDINATE']._serialized_start=36
  _globals['_COORDINATE']._serialized_end=70
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import optuna
import traceback

from . import study_as_df

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(prog="matsim-calibration", description="Calibration CLI")
    parser.add_argument('file', nargs=1, type=str, help="Path to input db")
    parser.add_argument("--name", type=str, default="calib", help="Calibration name")
    parser.add_argument("--output", default=None, help="Output path")
    args = parser.parse_args()

    study = optuna.load_study(
        study_name=args.name,
        storage="sqlite:///%s" % args.file[0],
    )

    if not args.output:
        args.output = args.file[0] + ".csv"

    df = study_as_df(study)
    df.to_csv(args.output, index=False)

    try:
        from .plot import plot_study
        plot_study(study)

    except ImportError:
        print("Could not plot study.")
        traceback.print_exc()




# -*- coding: utf-8 -*-

import os
import pandas as pd
import pyproj
import re
from collections import Counter
from shapely.geometry import box
from shapely.ops import transform

from .. import *

INPUT_FILES = 3


def is_format(f: os.DirEntry):
    return f.name.startswith("MiD2017_") and f.path.endswith(".csv")


def read_raw(household_file, person_file, trip_file):
    hh = pd.read_csv(household_file, delimiter=";", decimal=",",
                     quotechar="\"", low_memory=False, quoting=2)

    p = pd.read_csv(person_file, delimiter=";", decimal=",",
                    quotechar="\"", low_memory=False, quoting=2)

    t = pd.read_csv(trip_file, delimiter=";", decimal=",",
                    quotechar="\"", low_memory=False, quoting=2)

    return hh, p, t


def convert(data: tuple, regio=None):
    hh, pp, tt = data

    hhs = []
    for h in hh.itertuples():
        hh_id = str(int(h.H_ID_Lok))
        hhs.append(Household(hh_id, h.H_GEW, int(h.hhgr_gr2), int(h.anzauto_gr3), int(h.pedrad), int(h.motmop),
                             ParkingPosition.NA,
                             Mid2017.economic_status(h), Mid2017.household_type(h), 0, "",
                             geom=Mid2017.geom(h)))

    ps = []
    for p in pp.itertuples():
        p_id = str(int(p.HP_ID_Lok))
        hh_id = str(int(p.H_ID_Lok))

        ps.append(Person(
            p_id, p.P_GEW, hh_id, Mid2017.age(p), Mid2017.gender(p), Mid2017.employment(p), None,
            Mid2017.driving_license(p), Mid2017.car_avail(p), Mid2017.bike_avail(p), Mid2017.pt_abo_avail(p),
            Mid2017.mobile(p), Mid2017.present_on_day(p), int(p.ST_WOTAG), int(p.anzwege1)
        ))

    ts = []

    counts = Counter()

    for t in tt.itertuples():

        # business trips are not accounted
        if t.W_RBW == 1.0:
            continue

        # Person might have multiple days
        p_id = str(int(t.HP_ID_Lok))
        t_id = p_id + "%d%d" % (t.ST_WOCHE, t.ST_WOTAG)

        n = counts[t_id]
        counts[t_id] += 1

        depature = Mid2017.parse_time(t.W_SZ)

        trip = Trip(
            t_id + "_" + str(n),
            t.W_GEW, p_id, str(int(t.H_ID_Lok)),
            n, int(t.ST_WOTAG),depature, int(t.wegmin), float(t.wegkm),
            Mid2017.main_mode(t), Mid2017.purpose(t), None,
            float(t.wegkm) < 1000 and int(t.wegmin) < 1000 and depature is not None
        )

        if trip.purpose == Purpose.WAYBACK:
            if trip.n >= 3:
                # Use purpose of previous trip
                trip.purpose = ts[-2].purpose
            else:
                trip.purpose = Purpose.HOME

        ts.append(trip)

    return pd.DataFrame(hhs).set_index("hh_id"), pd.DataFrame(ps).set_index("p_id"), pd.DataFrame(ts).set_index("t_id")

def flip(x, y):
    """Flips the x and y coordinate values"""
    return y, x

class Mid2017:
    """ Converter for 2017 MID format """

    ct = pyproj.Transformer.from_proj(
        pyproj.Proj('epsg:3035'),  # source coordinate system
        pyproj.Proj('epsg:4326'),
        always_xy=True)  # destination coordinate system

    @staticmethod
    def coordinate(value, grid_size, scale):
        # EPSG:3035, coordinate is left lower point
        split = re.split(r"N|E", value)
        x = float(split[2]) * scale
        y = float(split[1]) * scale

        geom = box(x, y, x + grid_size, y + grid_size)

        return transform(Mid2017.ct.transform, geom).wkt

    @staticmethod
    def geom(h):
        if h.GITTER_500m.strip():
            return Mid2017.coordinate(h.GITTER_500m, 500, 100)
        elif h.GITTER_1km.strip():
            return Mid2017.coordinate(h.GITTER_1km, 1000, 1000)
        elif h.GITTER_5km.strip():
            return Mid2017.coordinate(h.GITTER_5km, 5000, 1000)

        return None

    @staticmethod
    def economic_status(h):
        return list(EconomicStatus)[int(h.oek_status) - 1]

    @staticmethod
    def household_type(h):
        x = int(h.hhtyp2)

        if x == 2:
            return HouseholdType.MULTI_W_CHILDREN
        elif x == 3:
            return HouseholdType.MULTI_WO_CHILDREN

        return HouseholdType.SINGLE if int(h.hhgr_gr2) == 1 else HouseholdType.MULTI_WO_CHILDREN

    @staticmethod
    def age(p):
        x = int(p.alter_gr)
        if x == 1:
            return 0
        elif x == 2:
            return 18
        elif x == 3:
            return 30
        elif x == 4:
            return 40
        elif x == 5:
            return 50
        elif x == 6:
            return 60
        elif x == 7:
            return 70
        elif x == 8:
            return 80

        return -1

    @staticmethod
    def gender(p):
        if p.HP_SEX == 1.0:
            return Gender.M
        elif p.HP_SEX == 2.0:
            return Gender.F

        return Gender.OTHER

    @staticmethod
    def employment(p):
        x = int(p.taet)
        if x == 1:
            return Employment.JOB_FULL_TIME
        elif x == 2:
            return Employment.STUDENT
        elif x == 3:
            return Employment.HOMEMAKER
        elif x == 4:
            return Employment.RETIREE

        return Employment.OTHER

    @staticmethod
    def driving_license(p):
        x = int(p.P_FSCHEIN)
        if x == 1:
            return Availability.YES
        elif x == 2:
            return Availability.NO

        return Availability.UNKNOWN

    @staticmethod
    def car_avail(p):
        x = int(p.P_STKFZ)
        if x == 1 or x == 2:
            return Availability.YES
        elif x == 3:
            return Availability.NO

        return Availability.UNKNOWN

    @staticmethod
    def bike_avail(p):
        x = int(p.vpedrad)
        if x == 1 or x == 2 or x == 3:
            return Availability.YES
        elif x == 4:
            return Availability.NO

        return Availability.UNKNOWN

    @staticmethod
    def pt_abo_avail(p):
        x = int(p.P_FKARTE)
        if x == 4 or x == 5:
            return Availability.YES
        elif x > 8:
            return Availability.UNKNOWN

        return Availability.NO

    @staticmethod
    def mobile(p):
        x = int(p.mobil)
        if x == 0:
            return False

        # potential null values are ignored, and true returned
        return True

    @staticmethod
    def present_on_day(p):
        x = int(p.P_STUM)
        if x == 1 or x == 2:
            return False

        return True

    @staticmethod
    def parse_time(x):
        t = x.strip().split(":")
        if len(t) < 2:
            return None

        return int(t[0]) * 60 + int(t[1])

    @staticmethod
    def main_mode(t):
        x = int(t.hvm)
        if x == 1:
            return TripMode.WALK
        elif x == 2:
            return TripMode.BIKE
        elif x == 3:
            return TripMode.RIDE
        elif x == 4:
            return TripMode.CAR
        elif x == 5:
            return TripMode.PT

        return TripMode.OTHER

    @staticmethod
    def purpose(t):
        x = int (t.zweck)
        if x == 1:
            return Purpose.WORK
        elif x == 2:
            return Purpose.WORK_BUSINESS
        elif x == 3:
            return Purpose.EDU
        elif x == 4:
            return Purpose.SHOPPING
        elif x == 5:
            return Purpose.PERSONAL_BUSINESS
        elif x == 6:
            return Purpose.TRANSPORT
        elif x == 7:
            return Purpose.LEISURE
        elif x == 8:
            return Purpose.HOME
        elif x == 9: # trip back to previous
            return Purpose.WAYBACK

        return Purpose.OTHER


""" This module contains dataclasses and methods for reading and processing survey data.
"""

__all__ = ["read_all", "ParkingPosition", "HouseholdType", "EconomicStatus", "Gender", "Employment", "Availability",
           "Purpose",
           "TripMode", "DistanceGroup", "DurationGroup", "SourceDestinationGroup",
           "Household", "Person", "Trip", "Activity"]

import os
from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Union, Tuple

import numpy as np
import pandas as pd


def _batch(iterable: list, max_batch_size: int):
    """ Batches an iterable into lists of given maximum size, yielding them one by one. """
    batch = []
    for element in iterable:
        batch.append(element)
        if len(batch) >= max_batch_size:
            yield batch
            batch = []
    if len(batch) > 0:
        yield batch


def read_all(dirs: Union[str, List[str]], regio=None) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """ Scan directories and read everything into one dataframe """

    from .formats import srv, mid

    hh = []
    pp = []
    tt = []

    # Allow to use without list
    if type(dirs) == str:
        dirs = [dirs]

    for d in dirs:

        for format in (srv, mid):

            files = []

            # Collect all files for each format
            for f in os.scandir(d):
                if not f.is_file():
                    continue
                if format.is_format(f):
                    files.append(f.path)

            files = sorted(files)

            if len(files) % format.INPUT_FILES != 0:
                print(files)
                raise ValueError("File structure is wrong. Need exactly %d files per region." % format.INPUT_FILES)

            for input_files in _batch(files, format.INPUT_FILES):
                print("Reading", format.__name__, *input_files)

                data = format.read_raw(*input_files)
                df = format.convert(data, regio)

                hh.append(df[0])
                pp.append(df[1])
                tt.append(df[2])

    hh = pd.concat(hh, axis=0)
    hh = hh[~hh.index.duplicated(keep='first')]
    print("Households: ", len(hh))

    pp = pd.concat(pp, axis=0)
    pp = pp[~pp.index.duplicated(keep='first')]
    pp.sort_index(inplace=True)
    print("Persons: ", len(pp))

    tt = pd.concat(tt, axis=0)
    tt = tt[~tt.index.duplicated(keep='first')]
    tt.sort_values(["p_id", "n"], inplace=True)
    print("Trips: ", len(tt))

    return hh, pp, tt


class StrEnum(str, Enum):
    pass


# this creates nice lowercase and JSON serializable names
# https://docs.python.org/3/library/enum.html#using-automatic-values
class AutoNameLower(StrEnum):
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()


class AutoNameLowerStrEnum(AutoNameLower):
    pass


class ParkingPosition(AutoNameLowerStrEnum):
    PRIVATE = auto()
    PUBLIC = auto()
    DIFFERENT = auto()
    NA = auto()


class HouseholdType(AutoNameLowerStrEnum):
    MULTI_W_CHILDREN = auto()
    MULTI_WO_CHILDREN = auto()
    SINGLE = auto()


class EconomicStatus(AutoNameLowerStrEnum):
    VERY_LOW = auto()
    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()
    VERY_HIGH = auto()
    UNKNOWN = auto()


class Gender(AutoNameLowerStrEnum):
    M = auto()
    F = auto()
    OTHER = auto()


class Employment(AutoNameLowerStrEnum):
    CHILD = auto()
    HOMEMAKER = auto()
    RETIREE = auto()
    UNEMPLOYED = auto()
    SCHOOL = auto()
    STUDENT = auto()
    TRAINEE = auto()

    JOB_FULL_TIME = auto()
    JOB_PART_TIME = auto()
    OTHER = auto()


class Availability(AutoNameLowerStrEnum):
    YES = auto()
    NO = auto()
    UNKNOWN = auto()


class Purpose(AutoNameLowerStrEnum):
    WORK = auto()
    WORK_BUSINESS = auto()
    EDU_KIGA = auto()
    """ Kinderkrippe/-garten  """
    EDU_PRIMARY = auto()
    """ Grundschule """
    EDU_SECONDARY = auto()
    """ Weiterführende Schule """
    EDU_HIGHER = auto()
    """ Berufs-, Fach-, Hochschule """
    EDU_OTHER = auto()
    """ Andere Bildungseinrichtung """
    EDU = auto()
    """ All Edu """

    SHOPPING = auto()
    """ General non differentiated shopping """

    SHOP_DAILY = auto()
    SHOP_OTHER = auto()
    PERSONAL_BUSINESS = auto()
    TRANSPORT = auto()
    LEISURE = auto()
    DINING = auto()
    OUTSIDE_RECREATION = auto()
    VISIT = auto()
    HOME = auto()
    WAYBACK = auto()
    OTHER = auto()


class TripMode(AutoNameLowerStrEnum):
    WALK = auto()
    BIKE = auto()
    CAR = auto()
    RIDE = auto()
    PT = auto()
    MOTORCYCLE = auto()
    OTHER = auto()


class DistanceGroup(AutoNameLowerStrEnum):
    """ These distance groups are designed so that they are roughly equally populated. """

    ZERO = auto()
    G_500M = auto()
    G_1KM = auto()
    G_2KM = auto()
    G_3KM = auto()
    G_5KM = auto()
    G_10KM = auto()
    G_25KM = auto()
    G_50KM = auto()
    G_100KM = auto()
    OVER_100KM = auto()

    @staticmethod
    def cut(values):
        bins = [0, 0.5, 1, 2, 3, 5, 10, 25, 50, 100]
        values = np.asarray(values)

        idx = np.digitize(values, bins, right=True)
        # Set ZERO group manually
        idx[np.where(values <= 0)] = 0
        return np.take(np.asarray(DistanceGroup, dtype=object), idx, axis=0)


class DurationGroup(AutoNameLowerStrEnum):
    """ Most common duration groups, right side is inclusive e.g <= 5 min """

    G_5MIN = auto()
    G_15MIN = auto()
    G_30MIN = auto()
    G_60MIN = auto()
    G_120MIN = auto()
    G_180MIN = auto()
    G_300MIN = auto()
    G_420MIN = auto()
    G_480MIN = auto()
    G_510MIN = auto()
    G_570MIN = auto()
    G_660MIN = auto()
    G_750MIN = auto()
    REST_OF_DAY = auto()

    @staticmethod
    def cut(values):
        bins = [5, 15, 30, 60, 120, 180, 300, 420, 480, 510, 570, 660, 750]

        values = np.asarray(values)
        idx = np.digitize(values, bins, right=True)
        return np.take(np.asarray(DurationGroup, dtype=object), idx, axis=0)


class SourceDestinationGroup(AutoNameLowerStrEnum):
    HOME_WORK = auto()
    HOME_CHILDCARE = auto()
    HOME_EDU = auto()
    HOME_BUSINESS = auto()
    HOME_SHOP = auto()
    HOME_LEISURE = auto()
    HOME_OTHER = auto()
    WORK_HOME = auto()
    CHILDCARE_HOME = auto()
    EDU_HOME = auto()
    BUSINESS_HOME = auto()
    SHOP_HOME = auto()
    LEISURE_HOME = auto()
    OTHER_HOME = auto()
    OTHER_WORK = auto()
    WORK_OTHER = auto()
    OTHER_OTHER = auto()

    UNKNOWN = auto()

    def source(self):
        if self.name.startswith("HOME"):
            return Purpose.HOME
        elif self.name.startswith("WORK"):
            return Purpose.WORK

        return Purpose.OTHER


@dataclass
class Household:
    """ Universal definition of household attributes """
    hh_id: str
    h_weight: float
    n_persons: int
    n_cars: int
    n_bikes: int
    n_other_vehicles: int
    car_parking: ParkingPosition
    economic_status: EconomicStatus
    type: HouseholdType
    region_type: int
    location: str
    zone: str = None
    """ A detailed zone, which can be more accurate than location. """
    income: float = None
    geom: object = None


@dataclass
class Person:
    """ Universal definition of person attributes."""
    p_id: str
    p_weight: float
    hh_id: str
    age: int
    gender: Gender
    employment: Employment
    restricted_mobility: bool
    driving_license: Availability
    car_avail: Availability
    bike_avail: Availability
    pt_abo_avail: Availability
    mobile_on_day: bool
    present_on_day: bool
    reporting_day: int
    n_trips: int


@dataclass
class Trip:
    """ Universal definition of trip attributes"""
    t_id: str
    t_weight: float
    p_id: str
    hh_id: str
    n: int
    day_of_week: int
    departure: int
    duration: int
    gis_length: float
    main_mode: TripMode
    purpose: Purpose
    sd_group: SourceDestinationGroup
    valid: bool
    from_location: str = pd.NA
    from_zone: str = pd.NA
    to_location: str = pd.NA
    to_zone: str = pd.NA


@dataclass
class Activity:
    """ Activity information (including leg) """
    a_id: str
    p_id: str
    n: int
    type: Purpose
    duration: int
    leg_dist: float
    leg_duration: float
    leg_mode: TripMode



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: matsim/pb/Wireformat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1amatsim/pb/Wireformat.proto\x12\tmatsim.pb\"L\n\x0cPBFileHeader\x12\x0f\n\x07version\x18\x01 \x01(\r\x12+\n\x0b\x63ontentType\x18\x02 \x01(\x0e\x32\x16.matsim.pb.ContentType*&\n\x0b\x43ontentType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x45VENTS\x10\x01\x42\x1c\n\x18org.matsim.core.utils.pbP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'matsim.pb.Wireformat_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030org.matsim.core.utils.pbP\001'
  _globals['_CONTENTTYPE']._serialized_start=119
  _globals['_CONTENTTYPE']._serialized_end=157
  _globals['_PBFILEHEADER']._serialized_start=41
  _globals['_PBFILEHEADER']._serialized_end=117
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

import math
import os
from abc import ABC, abstractmethod
from typing import Union, Sequence, Dict, Tuple, Callable

import optuna
import pandas as pd

# Type alias for input variables
CalibrationInput = Union[str, os.PathLike, dict, pd.DataFrame]


def to_float(x):
    return float(x.iloc[0]) if isinstance(x, pd.Series) else float(x)


def sanitize(x):
    if 'main_mode' in x.columns and 'mode' not in x.columns:
        x.rename(columns={'main_mode': 'mode'}, inplace=True)

    return x


class CalibratorBase(ABC):
    """ Base class for calibrators.  """

    def __init__(self,
                 modes: Sequence[str],
                 initial: CalibrationInput,
                 target: CalibrationInput,
                 lr: Callable[[int, str, float, optuna.Trial, optuna.Study], float] = None,
                 constraints: Dict[str, Callable[[str, float], float]] = None):
        """Abstract constructors for all calibrations. Usually the same parameters should be made available subclasses.

        :param modes: list of all relevant modes
        :param initial: dict of initial asc values
        :param target: target shares as csv or dict
        :param lr: learning rate schedule, will be called with (trial number, mode, update, trial, study)
        :param constraints: constraints for each mode, must return asc and will be called with original asc
        """
        self.modes = modes
        self.initial = sanitize(self.convert_input(initial))
        self.target = sanitize(self.convert_input(target))
        self.lr = lr
        self.constraints = constraints

    @property
    @abstractmethod
    def name(self):
        """ Short name of the calibrator. Will be used to differentiate attributes"""

        raise NotImplemented

    @property
    def num_targets(self):
        """ Number of targets/errors that will be returned in calc stats"""
        return len(self.target)

    @abstractmethod
    def init_study(self, study: optuna.Study):
        """ Initialize study instance  """
        raise NotImplemented

    @abstractmethod
    def update_config(self, study: optuna.Study, trial: optuna.Trial, prefix: str, config: dict):
        """ Calculate updated config for this trial. This method must interact with the trial using
         `trial.suggest_float` and by using the given prefix . """
        raise NotImplemented

    @abstractmethod
    def sample_initial(self, param: str) -> float:
        """  Sample initial value for parameter """
        raise NotImplemented

    @abstractmethod
    def update_step(self, param: str, last_trial: optuna.Trial, completed: Sequence[optuna.Trial]) -> float:
        """ Return update step for param based on last trial """
        raise NotImplemented

    @abstractmethod
    def calc_stats(self, trial: optuna.Trial, run_dir: str,
                   transform_persons: Callable = None,
                   transform_trips: Callable = None) -> Tuple[float]:
        """ Calculate and store statistics for a trial. These usually needs to be accessed for calculating updates.

         :return: reported error metrics
         """
        raise NotImplemented

    @classmethod
    def convert_input(cls, arg) -> pd.DataFrame:
        """ Method call to create target values from input argument """
        if isinstance(arg, pd.DataFrame):
            return arg
        if isinstance(arg, dict):
            # Convert simple dict of modes and values
            return pd.DataFrame(index=arg.keys(), data=arg.values(), columns=["value"]).rename_axis(index="mode")

        return pd.read_csv(arg)

    @staticmethod
    def get_mode_params(config: dict, mode: str):
        """ Retrieve the mode params from config object. Create if they don't exist yet. """
        if "planCalcScore" not in config:
            config["planCalcScore"] = {}

        if "scoringParameters" not in config["planCalcScore"]:
            config["planCalcScore"]["scoringParameters"] = []

        ps = config["planCalcScore"]["scoringParameters"]

        for p in ps:
            if "modeParams" in p:
                for m in p["modeParams"]:
                    if m["mode"] == mode:
                        return m

                m = {"mode": mode}
                p["modeParams"].append(m)
                return m

        m = {"mode": mode}
        ps.append({"modeParams": [m]})
        return m

    @staticmethod
    def calc_asc_update(z_i, m_i, z_0, m_0) -> float:
        """ Calculates the asc update for one step """
        # Update asc
        # (1) Let z_i be the observed share of mode i. (real data, to be reproduced)
        # (2) Run the simulation to convergence. Obtain simulated mode shares m_i.
        # (3) Do nothing for mode 0. For all other modes: add [ln(z_i) - ln(m_i)] – [ln(z_0) - ln(m_0)] to its ASC.
        # (4) Goto 2.
        if m_i == 0:
            return 0

        return math.log(z_i) - math.log(m_i) - (math.log(z_0) - math.log(m_0))



import gzip
import pathlib

import pytest

from matsim import Plans
from matsim import writers

HERE = pathlib.Path(__file__).parent

files = ['initial_plans.xml.gz']


@pytest.mark.parametrize('filepath', files)
def test_plan_writer(filepath):
    plans = Plans.plan_reader(HERE / filepath)

    with gzip.open("out_"+filepath, 'wb+') as f_write:
        writer = writers.PopulationWriter(f_write)
        writer.start_population(attributes={"coordinateReferenceSystem": "GK4"})

        for person, plan in plans:
            id = person.attrib['id']
            writer.start_person(id)
            writer.start_plan(selected=plan.attrib['selected'] == 'yes')
            activities = [e for e in plan if e.tag == 'activity']
            legs = [e for e in plan if e.tag == 'leg']
            for act, leg in zip(activities[:-1], legs):
                writer.add_activity(
                    type=act.attrib['type'], x=act.attrib['x'], y=act.attrib['y'],
                    end_time=time(act.attrib['end_time']))
                writer.add_leg(mode='walk' if leg.attrib['mode'] == 'walk_main' else leg.attrib['mode'])
            writer.add_activity(
                type=activities[-1].attrib['type'], x=activities[-1].attrib['x'], y=activities[-1].attrib['y'],
                end_time=time(activities[-1].attrib['end_time']))
            writer.end_plan()
            writer.end_person()

        writer.end_population()

    with gzip.open("out_" + filepath) as f_orig:
        with gzip.open(HERE / filepath) as f_new:
            assert f_orig.readlines() == f_new.readlines(), "input and output files don't match!"


def time(x):
    return sum(int(t)*f for t, f in zip(x.split(':'), (3600, 60, 1)))



""" Custom defined reference model for speed relative target """

# simple decision tree

# Inputs: length, (0)
# priority_lower, priority_equal, priority_higher,  (1,2,3)
# is_secondary_or_higher, is_primary_or_higher, is_motorway (4,5,6)
# speed (7)

params = [0.9, 0.8, 0.9, 0.8, 0.7, 0.7, 0.7, 0.7] * 3

def score(params, inputs):
    # lower prio
    if inputs[1] == 1:
        # highway
        if inputs[6] == 1:
            # >= 100kmh
            if inputs[7] >= 27.5:
                return params[0]
            else:
                return params[1]
        # primary
        elif inputs[5] == 1:
            # >= 80khm
            if inputs[7] >= 22:
                return params[2]
            else:
                return params[3]
        # secondary
        elif inputs[4] == 1:
            # 30kmh or lower
            if inputs[7] <= 10:
                return params[4]
            else:
                return params[5]
        else:
            # 30kmh or lower
            if inputs[7] <= 10:
                return params[6]
            else:
                return params[7]

    # equal prio
    elif inputs[2] == 1:
        # highway
        if inputs[6] == 1:
            # >= 100kmh
            if inputs[7] >= 27.5:
                return params[8]
            else:
                return params[9]
        # primary
        elif inputs[5] == 1:
            # >= 80khm
            if inputs[7] >= 22:
                return params[10]
            else:
                return params[11]
        # secondary
        elif inputs[4] == 1:
            # 30kmh or lower
            if inputs[7] <= 10:
                return params[12]
            else:
                return params[13]
        else:
            # 30kmh or lower
            if inputs[7] <= 10:
                return params[14]
            else:
                return params[15]

    # higher prio
    else:
        # highway
        if inputs[6] == 1:
            # >= 100kmh
            if inputs[7] >= 27.5:
                return params[16]
            else:
                return params[17]
        # primary
        elif inputs[5] == 1:
            # >= 80khm
            if inputs[7] >= 22:
                return params[18]
            else:
                return params[19]
        # secondary
        elif inputs[4] == 1:
            # 30kmh or lower
            if inputs[7] <= 10:
                return params[20]
            else:
                return params[21]
        else:
            # 30kmh or lower
            if inputs[7] <= 10:
                return params[22]
            else:
                return params[23]


def batch_loss(params, inputs, targets):
    error = 0
    for x, y in zip(inputs, targets):
        preds = score(params, x)
        error += (preds - y) ** 2
    return error



# -*- coding: utf-8 -*-

import xopen
import xml.etree.ElementTree as ET
import pandas as pd


class Network:

    _crsTag = 'coordinateReferenceSystem'

    def __init__(self, nodes, links, node_attrs, link_attrs, net_attrs=None):
        self.nodes = nodes
        self.links = links
        self.link_attrs = link_attrs
        self.node_attrs = node_attrs

        self.network_attrs = {}
        if net_attrs: self.network_attrs = net_attrs

    def __str__(self):
        return 'Network: {nodes} nodes, {links} links, {crs}'.format(
            nodes=len(self.nodes),
            links=len(self.links),
            crs=Network._crsTag in self.network_attrs and self.network_attrs[Network._crsTag] or 'No CRS')

    def as_geo(self, projection=None):
        import geopandas as gpd
        import shapely.geometry as shp

        """Return a GeoPandas GeoDataFrame containing link geometries suitable for plotting."""

        # Project the coords, if CRS is specified somehow
        if projection:
            crs = {'init': projection}
        elif Network._crsTag in self.network_attrs:
            crs = self.network_attrs[Network._crsTag]
        else:
            crs = None

        # attach xy to links
        full_net = (self.links
        .merge(self.nodes,
                left_on='from_node',
                right_on='node_id')
        .merge(self.nodes,
                left_on='to_node',
                right_on='node_id',
                suffixes=('_from_node', '_to_node'))
        )

        # create the geometry column from coordinates
        geometry = [shp.LineString([(ox,oy), (dx,dy)]) for ox, oy, dx, dy in zip(full_net.x_from_node, full_net.y_from_node, full_net.x_to_node, full_net.y_to_node)]

        # build the geopandas geodataframe
        geo_net = (gpd.GeoDataFrame(full_net,
            geometry=geometry,
            crs = crs)
            .drop(columns=['x_from_node','y_from_node','node_id_from_node','node_id_to_node','x_to_node','y_to_node'])
            )

        return geo_net

def read_network(filename, skip_attributes=False):
    """Read a MATSim network.xml.gz file. Returns a Network object with dataframes
    for nodes, links, node_attributes, and link_attributes. If the network has a CRS
    projection set, it will be available in network_attrs."""
    tree = ET.iterparse(xopen.xopen(filename, 'r'), events=['start', 'end'])
    nodes = []
    links = []
    node_attrs = []
    link_attrs = []

    network_attrs = {}

    attributes = node_attrs
    attr_label = 'node_id'
    current_id = None

    for xml_event, elem in tree:
        # the nodes element CLOSES at the end of the nodes, followed by links:
        if elem.tag == 'links' and xml_event == 'start':
            attributes = link_attrs
            attr_label = 'link_id'

        elif elem.tag == 'node' and xml_event == 'start':
            atts = elem.attrib
            current_id = atts['id']

            atts['node_id'] = atts.pop('id')
            atts['x'] = float(atts['x'])
            atts['y'] = float(atts['y'])
            if 'z' in atts: atts['z'] = float(atts['z'])

            nodes.append(atts)

        elif elem.tag == 'link' and xml_event == 'start':
            atts = elem.attrib
            current_id = atts['id']

            atts['link_id'] = atts.pop('id')
            atts['from_node'] = atts.pop('from')
            atts['to_node'] = atts.pop('to')

            atts['length'] = float(atts['length'])
            atts['freespeed'] = float(atts['freespeed'])
            atts['capacity'] = float(atts['capacity'])
            atts['permlanes'] = float(atts['permlanes'])

            if 'volume' in atts: atts['volume'] = float(atts['volume'])

            links.append(atts)


        elif elem.tag == 'attribute' and xml_event == 'end':
            if elem.attrib['name'] == Network._crsTag:
                network_attrs[Network._crsTag] = elem.text

            elif not skip_attributes:
                atts = {}
                atts[attr_label] = current_id
                atts['name'] = elem.attrib['name']
                atts['value'] = elem.text

                # TODO: pandas will make the value column "object" since we're mixing types
                if 'class' in elem.attrib:
                    if elem.attrib['class'] == 'java.lang.Long':
                        atts['value'] = float(elem.text)
                    if elem.attrib['class'] == 'java.lang.Integer':
                        atts['value'] = int(elem.text)

                attributes.append(atts)

        # clear the element when we're done, to keep memory usage low
        if elem.tag in ['node', 'link'] and xml_event == 'end':
            elem.clear()

    nodes = pd.DataFrame.from_records(nodes)
    links = pd.DataFrame.from_records(links)
    node_attrs = pd.DataFrame.from_records(node_attrs)
    link_attrs = pd.DataFrame.from_records(link_attrs)

    return Network(nodes, links, node_attrs, link_attrs, network_attrs)
